
//**********************************************************************
//* ibmasm.c                                                           *
//*                                                                    *
//* Written by Pete Reynolds                                           *
//* (C) Copyright IBM Corporation, 1998-2007                           *
//*                                                                    *
//*  This software may be used and distributed according to the terms  *
//*  of the GNU Public License, incorporated herein by reference.      *
//**********************************************************************

//**********************************************************************
//* ibmasm.c - This source file implements all Service Processor USB   *
//*            device driver interface routines. It requires that the  *
//*            libusb library be available at build time.              *
//*                                                                    *
//* External library functions required from libusb                    *
//* -----------------------------------------------                    *
//* usb_init              -  Initialize libusb                         *
//* usb_find_busses       -  Find's all USB busses on system	       *
//* usb_find_devices      -  Find all devices on all USB devices       *
//* usb_open              -  Opens a USB device                        *
//* usb_close             -  Closes a USB device                       *
//* usb_set_configuration -  Sets the active configuration of a device *
//* usb_claim_interface   -  Claim an interface of a device            *
//* usb_release_interface -  Releases a previously claimed interface   *
//* usb_bulk_write        -  Write data to a bulk endpoint             *
//* usb_bulk_read         -  Read data from a bulk endpoint            *
//*                                                                    *
//*                                                                    *
//* Change History                                                     *
//* ----------------------------------------------------------------   *
//*   Date   - Change ID  - Description                                *
//* ----------------------------------------------------------------   *
//* 11/18/02 - PHR_208637 - Initial code added to source library       *
//* ----------------------------------------------------------------   *
//* 07/31/03 - PHR_220505 - Add multi-node support                     *
//* ----------------------------------------------------------------   *
//* 09/12/03 - PHR_223662 - use signal() instead of sigaction()        *
//* ----------------------------------------------------------------   *
//* 10/01/03 - PHR_225158 - Add 5.2.16 reset event support             *
//* ----------------------------------------------------------------   *
//* 06/10/04 - T. Yu      - Add debug switch for turning on/off debug  *
//*                         messages                                   *
//* ----------------------------------------------------------------   *   
//* 06/21/04 - DDH        - Reload ibmasm if no SPs found 248179       *
//* ----------------------------------------------------------------   *
//* 06/23/04 - T. Yu      - stop retries when sp reset to prevent from *
//*                         errorcode 8 returned to lflash             *   
//**********************************************************************
#include "ibmasm.h"

// *************************
// * GLOBAL DECLARATIONS   *
// *************************
static int ResetUSB;
static int CommandReceived;
static int InterruptedBySignal;


// *************************
// * LOCAL SIGNAL HANDLERS *
// *************************
// Start of PHR_223662
void sig_handler(int signum)
{
  switch (signum)
  {
     case SIGTERM:
     case SIGINT:
        InterruptedBySignal = TRUE;
     break;

     case SIGUSR1:
        CommandReceived = TRUE;
     break;

     case SIGUSR2:
        ResetUSB = TRUE;
        break;

  }

}
// End of PHR_223662


// *******************************
// * LOCAL FUNCTION PROTOTYPES   *
// *******************************
void InitSharedMem(PSP_SHARED_MEM pSpSharedMem);

int  InitDynamicLinks(PSP_SHARED_MEM pSpSharedMem);     // PHR_LINK

int  CountSPs(PSP_SHARED_MEM pSpSharedMem);             // PHR_LINK

int  InitNode(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pTempDevNode, int node);   // PHR_LINK

int  SendInitCmds(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode);

int  GetMacAddr(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode);

int  AddNode(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pTempDevNode);

int  GetNodeFromPid(PSP_SHARED_MEM pSpSharedMem);

int  ScanRespPkt(PSP_SHARED_MEM pSpSharedMem, unsigned char *pBuffer);

int  ScanEventPkt(PSP_SHARED_MEM pSpSharedMem, int node, unsigned char *EventBuff);  // PHR_225158

int  CreateLockFile(void);

int  PollForEvents(PSP_SHARED_MEM pSpSharedMem, int node);

int  EventHandler(PSP_SHARED_MEM pSpSharedMem);

int  CommandHandler(PSP_SHARED_MEM pSpSharedMem);

int  LockSharedMemory(void);

void UnlockSharedMemory(int FileDesc);

void DeleteLockFile(void);

void WakeEventTasks(PSP_SHARED_MEM pSpSharedMem, int node);

int  SendOSisDown(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode);

int  SendCmdToSP(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pSpDevNode, unsigned char *pBuffer, int BuffLen, int Retrys );

int  EnableShutdownEvent(PSP_SHARED_MEM pSpSharedMem, int node);

unsigned char usb_cksum(unsigned char* buf, int size);

// added by PHR_225158
void process_reset_event(PSP_SHARED_MEM pSpSharedMem, int node, int event);

//************************************************************************
// main() - main allocates shared memory which is used by itself and the *
//          shared object (libsysSp.so) as a form of inter-process data  *
//          communication. It's main purpose is to receive dot command   *
//          requestes from an application (i.e. director agent) thru the *
//          libsysSp.so shared object and pass them on to the service    *
//          processor adapter via the USB library interface. Once the 	 *
//          command has been sent it waits for the response, copies that *
//          response to the shared memory and sets a flag informing the  *
//          shared object that the command response is available. 		 *
//                                                                       *
//          It is also spawns a child process that is responsible for    *
//          polling the service processor adapter for events. If an event*
//          is reported by the service processor the child process uses  *
//          the same method that the dot command handler uses to inform  *
//          the shared object that there is an event to process.         *
//                                                                       *
//          If a USB disconnect or re-enumeration is detected ibmasm will*
//          re-load itself so that it can establish a new USB handle to  *
//          the re-enumerated USB interface. The shared object should    *
//          not be affected by this, it should simply wait until the 	 *
//          ibmasm reset is complete (which can take up to 15 secs) and  *
//          continue processing. 										 *
//                                                                       *
// Parameters:                                                           *
//                                                                       *
// Return Codes:                                                         *
//          EXIT_SUCCESS                                                 *
//          EXIT_FAILURE                                                 *
//                                                                       *
//************************************************************************
int main(int argc, char *argv[])	
{
   int rc;
   int IpcRC;
   int SpCount;
   int Logged;
   int NodesHandled;
   int node;
   int LockFileDesc;			
   int count_trys = 0;
   PSP_DEV_NODE pDevNode;

   // SHARED MEMORY DATA
   int MemID;
   void  *SharedMem = (void *)0;
   PSP_SHARED_MEM pSpSharedMem;

   SP_DEV_NODE   TempDevNode;
   PSP_DEV_NODE pTempDevNode = &TempDevNode;

   // FORK DATA
   pid_t ForkPid;

   // Start of PHR_223662
   int i;
   int NumSignals;

   /* pointer to a signal handler */
   typedef void (*sighandler_t)(int);
   sighandler_t sig;

   // SIGNAL HANDLER  -  cannot catch SIGKILL or SIGSTOP
   int SigArray[] = {SIGINT, SIGTERM, SIGUSR1, SIGUSR2};

   NumSignals = (sizeof(SigArray) / sizeof(SigArray[0]));

   // Enable signal handling
   for (i = 0; i < NumSignals; i++)
   {
       sig = signal(SigArray[i], sig_handler);

       if (sig == SIG_ERR)
       {
          if(debug) syslog(LOG_CRIT, "signal() returned SIG_ERR, unable to register for signal. %i.\n", SigArray[i]);
          return EXIT_FAILURE;
       }
   }
   // End of PHR_223662


   // ENSURE THAT IBMASM IS NOT ALREADY LOADED
   rc = CreateLockFile();

   if (rc != RC_SUCCESS)
   {
       syslog(LOG_CRIT, "Lock file exists, IBMASM may already be loaded, returning EXIT_FAILURE.\n");
       exit(EXIT_FAILURE);
   }


   unlink(MEM_LOCK_FILE);    // PHR_TODO   not sure about this
   sleep(2);

   // INITIALIZE GLOBALS
   InterruptedBySignal = FALSE;
   CommandReceived     = FALSE;
   ResetUSB            = FALSE;

   Logged = FALSE;

   // ALLOCATE SHARED MEMORY
   MemID = shmget( (key_t)SHARED_MEM_KEY, sizeof(SP_SHARED_MEM), 0666 | IPC_CREAT);

   if(debug) syslog(LOG_CRIT, "alloc shared mem; sizes = %x, %x, %x, %x, %x.\n",
                     sizeof(SP_SHARED_MEM), sizeof(pid_t), sizeof(SP_DEV_NODE),
		     sizeof(SP_CALLER_LIST), sizeof(SP_HANDLE_LIST) );
  

   if ( MemID == IPC_FAILURE )
   {
	   if(debug) syslog(LOG_CRIT, "shmget() failed.\n");
	   return RC_SYSTEM_ERROR;
   }

   // ATTACH THIS PROCESS TO THE SHARED MEMORY
   SharedMem = shmat(MemID, (void *)0, 0);

   if (SharedMem == (void *)IPC_FAILURE )
   {
	   if(debug) syslog(LOG_CRIT, "shmat() failed.\n");
	   return RC_SYSTEM_ERROR;
   }

   // SET UP THE LOCAL POINTER TO SHARED MEMORY
   pSpSharedMem = (PSP_SHARED_MEM)SharedMem;


   // LOCK MEMORY
   LockFileDesc = LockSharedMemory();

   if (LockFileDesc == IPC_FAILURE)
   {
       if(debug) syslog(LOG_CRIT, "Initialization phase could not lock shared memory, returning EXIT_FAILURE.\n");

	   // DELETE SHARED MEMORY
       shmdt(SharedMem);
	   shmctl(MemID, IPC_RMID, 0);

       DeleteLockFile();
       exit(EXIT_FAILURE);
   }
   //if(debug) syslog(LOG_CRIT, "Main() Locked shared memory.\n");


   // PHR_TODO, TRY TO DYNAMICALLY ADD AND REMOVE NODES INSTEAD OF RE-LOADING IBMASM
   if ( argc == 1)
   {
       // INITIALIZE THE SHARED MEMORY
       InitSharedMem(pSpSharedMem);
   }
   else
   {
       kill(pSpSharedMem->IbmasmsPid, SIGKILL);

       if (pSpSharedMem->ChildsPid != -1) 
       {
          kill(pSpSharedMem->ChildsPid, SIGKILL);
       }
       pSpSharedMem->InitialLoad = FALSE;
       pSpSharedMem->SpCount     = 0;
       pSpSharedMem->IbmasmsPid  = getpid();

       for (node = 0; node < MAX_SP_NODES; node++)
       {
          pSpSharedMem->SPDevNode[node].pUsbHandle = (void *)INVALID_USB_HANDLE;
       }
       if(debug) syslog(LOG_CRIT, "Sleep 20 seconds to allow node reset to complete.\n");
       sleep(20); // Sleep long enough to allow resetting nodes to finish.
   }


   // LINK TO THE USB LIBRARY
   rc = InitDynamicLinks(pSpSharedMem);

   if (rc != RC_SUCCESS)
   {
       if(debug) syslog(LOG_CRIT, "Initialization phase could not link libusb.so, returning EXIT_FAILURE.\n");

       UnlockSharedMemory(LockFileDesc);	

	   // DELETE SHARED MEMORY
       shmdt(SharedMem);
	   shmctl(MemID, IPC_RMID, 0);

       DeleteLockFile();
       exit(EXIT_FAILURE);
   }

   UnlockSharedMemory(LockFileDesc);	
   if(debug) syslog(LOG_CRIT, "Main() Initial SPCount.\n");
   
   // FIND OUT HOW MANY SPs ARE AVAILABLE TO THE SYSTEM
   SpCount = 0;

   while(!SpCount)
   {
       //syslog(LOG_CRIT, "Enter SpCount loop.\n");

       if (InterruptedBySignal)
       {
           syslog(LOG_CRIT, "Did not find an RSA II device in the system, returning EXIT_SUCCESS.\n");

           dlclose(pSpSharedMem->pLib);

		   shmdt(SharedMem);
           shmctl(MemID, IPC_RMID, 0);

		   DeleteLockFile();

           exit(EXIT_SUCCESS);
       }

        //syslog(LOG_CRIT, "Check trys.\n");

       if (( count_trys > 0 ) && ( SpCount == 0 )) 
       {
          pSpSharedMem->Resetting = TRUE;
          dlclose(pSpSharedMem->pLib);
          DeleteLockFile();
          if(debug) syslog(LOG_CRIT, "Main() exiting for 0 count.\n");
          system("/sbin/ibmasm 1");
          exit(EXIT_SUCCESS);
       }
       else
       {
          count_trys++;
       }
        //syslog(LOG_CRIT, "Lock mem.\n");

       // LOCK MEMORY
       LockFileDesc = LockSharedMemory();

       if (LockFileDesc == IPC_FAILURE)
       {
           if(debug) syslog(LOG_CRIT, "Initialization phase could not lock shared memory, returning EXIT_FAILURE.\n");

           dlclose(pSpSharedMem->pLib);

           shmdt(SharedMem);

           shmctl(MemID, IPC_RMID, 0);

           DeleteLockFile();
           exit(EXIT_FAILURE);
       }
        //syslog(LOG_CRIT, "Count Sps.\n");

       SpCount = CountSPs(pSpSharedMem);

       if(debug) syslog(LOG_CRIT, "Found %i RSA II node(s) in the system.\n", SpCount);

       if(SpCount)
       {
           //syslog(LOG_CRIT, "non zero Count Sps.\n");
           // MAKE SURE THE ADAPTER IS ENUMERATED BEFORE INITIALIZING OR USB READ/WRITE
           // ERRORS WILL OCCUR. THESE SEEM TO HAPPEN ANYWAY NO MATTER HOW LONG WE WAIT???
           //sleep(30);       // PHR_225158
           sleep(5);          // PHR_225158

           NodesHandled = 0;

           for (node = 0; node < MAX_SP_NODES; node++)
	        {
	           rc = InitNode(pSpSharedMem, pTempDevNode, node);           // PHR_LINK

	           if (rc != RC_SUCCESS)
              {
                   if(debug) syslog(LOG_CRIT, "InitNode failed, SpCount = 0.\n");
                   SpCount = 0;
		             break;
	           }

              rc = SendInitCmds(pSpSharedMem, pTempDevNode);

	           if (rc != RC_SUCCESS)
              {
                   if(debug) syslog(LOG_CRIT, "SendInitCmds failed, freeing USB stuff, SpCount = 0.\n");
		             if(debug) syslog(LOG_CRIT, "release interface handle = %x, interface = %x.\n", pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );
                   pSpSharedMem->sp_usb_release_interface(pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );       // PHR_LINK
                   pSpSharedMem->sp_usb_close(pTempDevNode->pUsbHandle);                                      // PHR_LINK
                   SpCount = 0;
		             break;
	           }

              rc = GetMacAddr(pSpSharedMem, pTempDevNode);

              if (rc != RC_SUCCESS)
              {
                   if(debug) syslog(LOG_CRIT, "GetMacAddr failed, freeing USB stuff, SpCount = 0.\n");
		             if(debug) syslog(LOG_CRIT, "release interface handle = %x, interface = %x.\n", pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );
                   pSpSharedMem->sp_usb_release_interface(pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );       // PHR_LINK
                   pSpSharedMem->sp_usb_close(pTempDevNode->pUsbHandle);                                      // PHR_LINK
                   SpCount = 0;
		             break;
	           }

              rc = GetNodeMode(pSpSharedMem, pTempDevNode);

              if (rc != RC_SUCCESS)
              {
                   if(debug) syslog(LOG_CRIT, "GetNodeMode failed, freeing USB stuff, SpCount = 0.\n");
		             if(debug) syslog(LOG_CRIT, "release interface handle = %x, interface = %x.\n", pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );
                   pSpSharedMem->sp_usb_release_interface(pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );       // PHR_LINK
                   pSpSharedMem->sp_usb_close(pTempDevNode->pUsbHandle);                                      // PHR_LINK
                   SpCount = 0;
		             break;
	           }

              rc = AddNode(pSpSharedMem, pTempDevNode);

              if (rc != RC_SUCCESS)
              {
                   if(debug) syslog(LOG_CRIT, "AddNode failed, freeing USB stuff, SpCount = 0.\n");
		             if(debug) syslog(LOG_CRIT, "release interface handle = %x, interface = %x.\n", pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );
                   pSpSharedMem->sp_usb_release_interface(pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );       // PHR_LINK
                   pSpSharedMem->sp_usb_close(pTempDevNode->pUsbHandle);                                      // PHR_LINK
                   SpCount = 0;
		             break;
	           }

              NodesHandled += 1;

              if ( (NodesHandled >= SpCount) && (pSpSharedMem->SPDevNode[0].State != NODE_AVAILABLE) )
	           {
	               break;
	           }
           }

           // DON'T KNOW IF THIS CAN HAPPEN BUT IF THERE ARE RSA II DEVICES IN THEY SYSTEM AND
           // THEY FAIL TO INITIALIZE, LOG IT
           if(!SpCount)
	        {
               for (node = 0; node < MAX_SP_NODES; node++)
               {
                  if ( pSpSharedMem->SPDevNode[node].pUsbHandle != (void *)INVALID_USB_HANDLE )
                  {
                     if(debug) syslog(LOG_CRIT, "main loop 1, no sp: release interface handle = %x, interface = %x.\n", pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface);
                     pSpSharedMem->sp_usb_release_interface(pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface );     // PHR_LINK
                     pSpSharedMem->sp_usb_close(pSpSharedMem->SPDevNode[node].pUsbHandle);                                    // PHR_LINK
                     pSpSharedMem->SPDevNode[node].pUsbHandle = (void *)INVALID_USB_HANDLE;
                     if (pSpSharedMem->SPDevNode[node].State == NODE_OPENED)
                     {
                        pSpSharedMem->SPDevNode[node].State = NODE_CLOSED;
                     }
                  }
               }
               if ( (!Logged) && (!pSpSharedMem->Resetting))
               {
                   Logged = TRUE;
                   syslog(LOG_CRIT, "RSA II device not found, will reload to search every 10 seconds forever.\n");
               }
           }
           else if (pSpSharedMem->SPDevNode[0].State != NODE_OPENED)
           {
               for (node = 1; node < MAX_SP_NODES; node++)
               {
                   if ( pSpSharedMem->SPDevNode[node].pUsbHandle != (void *)INVALID_USB_HANDLE )
                   {
                      if(debug) syslog(LOG_CRIT, "main loop 1, no pri: release interface handle = %x, interface = %x.\n", pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface);
		                pSpSharedMem->sp_usb_release_interface(pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface );     // PHR_LINK
                      pSpSharedMem->sp_usb_close(pSpSharedMem->SPDevNode[node].pUsbHandle);                                    // PHR_LINK
                      pSpSharedMem->SPDevNode[node].pUsbHandle = (void *)INVALID_USB_HANDLE;
                      if (pSpSharedMem->SPDevNode[node].State == NODE_OPENED)
                      {
                         pSpSharedMem->SPDevNode[node].State = NODE_CLOSED;
                      }
                   }
               }
               SpCount = 0;
               if ( (!Logged) && (!pSpSharedMem->Resetting))
               {
                  Logged = TRUE;
                  syslog(LOG_CRIT, "SP Primary Node not found, will reload to search every 10 seconds forever.\n");
               }
           }
       }       // end if SpCount
       else
       {
          if ( (!Logged) && (!pSpSharedMem->Resetting))
	       {
	           Logged = TRUE;
	           syslog(LOG_CRIT, "RSA II device not found, will reload to search every 10 seconds forever.\n");
	       }
       }

       UnlockSharedMemory(LockFileDesc);	
       if(debug) syslog(LOG_CRIT, "Main() unlocked shared memory.\n");

       if (SpCount == 0) sleep(10);
   }	                             // end while !SpCount


   // INITIALIZE SHARED MEMORY FLAGS
   pSpSharedMem->SpCount   = SpCount;
   
   if(debug) syslog(LOG_CRIT, "main() at init: pSpSharedMem->Resetting = FALSE;\n");
   pSpSharedMem->Resetting = FALSE;


   // DISPLAY WELCOME MESSAGE	FOR BUILD ibmusbasm-1.46-2.i386.rpm  Version 1.46
   if (pSpSharedMem->InitialLog)
   {
       pSpSharedMem->InitialLog = FALSE; // don't log "loaded" message more than once
         
       if (SpCount == 1)
       {
          syslog(LOG_CRIT, "V1.46 IBM RSA II 32-bit device support loaded.\n");
       }
       else
       {
          syslog(LOG_CRIT, "V1.46 IBM RSA II 32-bit device support loaded, %i USB SP device nodes found.\n", SpCount);
       }
   }

   if(debug) syslog(LOG_CRIT, "Starting Fork process - CommandHandler()\n");
   ForkPid = fork();

   switch (ForkPid)
   {
       case -1:	  // ERROR CONDITION
       {
           if(debug) syslog(LOG_CRIT, "Fork of child process failed, Returning EXIT_FAILURE\n");

		   IpcRC = shmdt(SharedMem);

		   if ( IpcRC == IPC_FAILURE)
		   {
			   if(debug) syslog(LOG_CRIT, "FreeResources() call to shmdt failed.\n");
			   rc = IPC_FAILURE;
		   }

		   shmctl(MemID, IPC_RMID, 0);

		   if (IpcRC == IPC_FAILURE)
		   {
			   if(debug) syslog(LOG_CRIT, "FreeResources() call to shmclt failed.\n");
			   rc = IPC_FAILURE;
		   }

		   DeleteLockFile();

           exit(EXIT_FAILURE);
       }
       case 0:	  // CHILD PROCESS
       {
           if(debug) syslog(LOG_CRIT, "Forking child process.\n");

           rc = EventHandler(pSpSharedMem);

           switch(rc)
	       {
	            case RC_INTERRUPTED:   // EXIT SUCCESS
	            {
	                //if(debug) syslog(LOG_CRIT, "EventHandler() was interrupted by a signal, returning EXIT_SUCCESS.\n");
                    rc = EXIT_SUCCESS;
	                break;
	            }
	            case RC_RESET:	      // SIGNAL THE COMMAND HANDLER AND EXIT SUCCESS
	            {
	                //if(debug) syslog(LOG_CRIT, "EventHandler sending reset to ibmasm's CommandHandler. \n");
                    // TELL IBMASM THAT THERE WAS A CHANGE IN NODE STATUS              // PHR_225158
                    kill( pSpSharedMem->IbmasmsPid, SIGUSR2);                          // PHR_225158
	                rc = EXIT_SUCCESS;
	                break;
                }
	            default:               // EXIT WITH AN ERROR
	            {
	                //if(debug) syslog(LOG_CRIT, "EventHandler() returned with an error, returning EXIT_FAILURE.\n");
                    rc = EXIT_FAILURE;
	                break;
	            }
           }      // end switch on rc

           break;    // exit child case
       }
       default:	     // PARENT PROCESS
       {
           if(debug) syslog(LOG_CRIT, "Loading parent process.\n");
           rc = CommandHandler(pSpSharedMem);

           switch(rc)
	        {
	           case RC_INTERRUPTED:   // CLEAN UP AND EXIT, KILL CHILD PROCESS
	           {
                  InterruptedBySignal = FALSE;

                  if(debug) syslog(LOG_CRIT, "Main() was interrupted by a signal, returning RC_SUCCESS.\n");
	               kill(pSpSharedMem->ChildsPid, SIGKILL);  // SIGTERM

	               // RELEASE THE USB INTERFACES AND CLOSE THE USB DEVICES
                  NodesHandled = 0;

	               for (node = 0; node < MAX_SP_NODES; node++)
                  {
                     if (pSpSharedMem->SPDevNode[node].State != NODE_AVAILABLE)
		               {
		                   NodesHandled+=1;

                         pDevNode = &pSpSharedMem->SPDevNode[node];

                         if (pSpSharedMem->SPDevNode[node].State == NODE_OPENED)
                         {
		                      rc = SendOSisDown(pSpSharedMem, pDevNode);
                         }

                         if ( pSpSharedMem->SPDevNode[node].pUsbHandle != (void *)INVALID_USB_HANDLE )
                         {
                            if(debug) syslog(LOG_CRIT, "main loop: release interface handle = %x, interface = %x.\n", pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface);
                            pSpSharedMem->sp_usb_release_interface(pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface );   // PHR_LINK
                            pSpSharedMem->sp_usb_close(pSpSharedMem->SPDevNode[node].pUsbHandle);                                  // PHR_LINK
                            pSpSharedMem->SPDevNode[node].pUsbHandle = (void *)INVALID_USB_HANDLE;
                            if (pSpSharedMem->SPDevNode[node].State == NODE_OPENED)
                            {
                               pSpSharedMem->SPDevNode[node].State = NODE_CLOSED;
                            }
                         }

                         //if (NodesHandled >= SpCount)
                         //{
                         //    break;
                         //}
		               }
		            }

                  dlclose(pSpSharedMem->pLib);

			         IpcRC = shmdt(SharedMem);

			         if (IpcRC == IPC_FAILURE)
			         {
			            //if(debug) syslog(LOG_CRIT, "FreeResources() call to shmdt failed.\n");
			         }

			         shmctl(MemID, IPC_RMID, 0);

			         if (IpcRC == IPC_FAILURE)
			         {
			            if(debug) syslog(LOG_CRIT, "FreeResources() call to shmclt failed.\n");
			   	        rc = IPC_FAILURE;
			         }

			         DeleteLockFile();

                  syslog(LOG_CRIT, "RSA II device support terminated. \n");
                  rc = EXIT_SUCCESS;

		            break;
	           }
	           case RC_RESET:	      // RELOAD IBMASM
              {
                   // Start of PHR_225158
                   if(debug) syslog(LOG_CRIT, "Lost connectivity to the SP, re-initializing ibmasm.\n");

                   InterruptedBySignal = FALSE;


				       if(debug) syslog(LOG_CRIT, "main() inside RC_RESET CASE: pSpSharedMem->Resetting = TRUE;\n");
                   pSpSharedMem->Resetting = TRUE;

                   // kill the child process
                   kill(pSpSharedMem->ChildsPid, SIGKILL);

                   NodesHandled = 0;

                   for (node = 0; node < MAX_SP_NODES; node++)
                   {
                       if (pSpSharedMem->SPDevNode[node].State != NODE_AVAILABLE)
                       {
                           switch (pSpSharedMem->SPDevNode[node].State)
                           {
                               case NODE_OPENED:
                               {
                                   pSpSharedMem->SPDevNode[node].State = NODE_CLOSED;
                                   if(debug) syslog(LOG_CRIT, "main loop 2: release interface handle = %x, interface = %x.\n", pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface);
			                          pSpSharedMem->sp_usb_release_interface(pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface );     // PHR_LINK
                                   pSpSharedMem->sp_usb_close(pSpSharedMem->SPDevNode[node].pUsbHandle);                                      
            				           break;
                               }
                               case NODE_DISCONNECTED:
                               {
                                   process_reset_event(pSpSharedMem, node, EVENT_DISCONNECT);
                                   syslog(LOG_CRIT, "RSA II device support suspended for node %i, connection lost.\n", pSpSharedMem->ResetEventInfo.node );
                                   if ( pSpSharedMem->SPDevNode[node].pUsbHandle != (void *)INVALID_USB_HANDLE )
                                   {
                                      if(debug) syslog(LOG_CRIT, "main loop 2: release interface handle = %x, interface = %x.\n", pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface);
                                      pSpSharedMem->sp_usb_release_interface(pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface );     // PHR_LINK
                                      pSpSharedMem->sp_usb_close(pSpSharedMem->SPDevNode[node].pUsbHandle);                                      
                                      pSpSharedMem->SPDevNode[node].pUsbHandle = (void *)INVALID_USB_HANDLE;
                                   }
                                   sleep(2);
                                   break;
                               }
                               case NODE_RESETTING:
                               {
                                   syslog(LOG_CRIT, "RSA II device support suspended for node %i, node is resetting.\n", pSpSharedMem->ResetEventInfo.node );
                                   if ( pSpSharedMem->SPDevNode[node].pUsbHandle != (void *)INVALID_USB_HANDLE )
                                   {
                                      if(debug) syslog(LOG_CRIT, "main loop 2: release interface handle = %x, interface = %x.\n", pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface);
                                      pSpSharedMem->sp_usb_release_interface(pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface );     // PHR_LINK
                                      pSpSharedMem->sp_usb_close(pSpSharedMem->SPDevNode[node].pUsbHandle);                                      
                                      pSpSharedMem->SPDevNode[node].pUsbHandle = (void *)INVALID_USB_HANDLE;
                                   }
                                   break;
                               }
                           }
                                 // PHR_LINK

                           NodesHandled+=1;

                           //if (NodesHandled >= SpCount)
                           //{
                           //    break;
                           //}
                       }
                   }
                   // End of PHR_225158

                   dlclose(pSpSharedMem->pLib);
                   DeleteLockFile();
                   system("/sbin/ibmasm 1");
                   rc = EXIT_SUCCESS;

                   break;
               }
	           default:              // CLEAN UP AND EXIT WITH AN ERROR
	           {
	               if(debug) syslog(LOG_CRIT, "Main() returned with a RC_SYSTEM_ERROR. \n");
                  InterruptedBySignal = FALSE;

	               // kill the child process
	               kill(pSpSharedMem->ChildsPid, SIGKILL);

	               // RELEASE THE USB INTERFACES AND CLOSE THE USB DEVICES
                  NodesHandled = 0;

		            for (node = 0; node < MAX_SP_NODES; node++)
		            {
                     if (pSpSharedMem->SPDevNode[node].State == NODE_OPENED)
		               {
                         pDevNode = &pSpSharedMem->SPDevNode[node];
			                rc = SendOSisDown(pSpSharedMem, pDevNode);
                         pSpSharedMem->SPDevNode[node].State = NODE_CLOSED;
                     }

                     if ( pSpSharedMem->SPDevNode[node].pUsbHandle != (void *)INVALID_USB_HANDLE )
                     {
                        if(debug) syslog(LOG_CRIT, "main loop 3: release interface handle = %x, interface = %x.\n", pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface);
                        pSpSharedMem->sp_usb_release_interface(pSpSharedMem->SPDevNode[node].pUsbHandle, pSpSharedMem->SPDevNode[node].UsbInterface );     // PHR_LINK
                        pSpSharedMem->sp_usb_close(pSpSharedMem->SPDevNode[node].pUsbHandle);                                      
                        pSpSharedMem->SPDevNode[node].pUsbHandle = (void *)INVALID_USB_HANDLE;
                     }

	               }

                  dlclose(pSpSharedMem->pLib);

				      IpcRC = shmdt(SharedMem);
                 
				      if (IpcRC == IPC_FAILURE)
				      {
				          if(debug) syslog(LOG_CRIT, "FreeResources() call to shmdt failed.\n");
				      }
                 
				      shmctl(MemID, IPC_RMID, 0);
                 
				      if (IpcRC == IPC_FAILURE)
				      {
				          if(debug) syslog(LOG_CRIT, "FreeResources() call to shmclt failed.\n");
				      	   rc = IPC_FAILURE;
				      }
                 
				      DeleteLockFile();
                 
                  syslog(LOG_CRIT, "RSA II device support terminated. \n");
	               rc = EXIT_FAILURE;
                 
	               break;
	           }
           }       // end switch on rc

           break;  // break out of default case - Parent process
       }           // end default - Parent process
   }               // end switch on fork

   exit(rc);
}                  // end main




// ***********************************************************************
// * CommandHandler() - This routine passes all user requests destined   *
// *                    for the adapter to the adpater via USB bulk read *
// *                    and write library calls. It validates all the    *
// *                    callers parameters and registration data then    *
// *                    calls sendCmdToSP to make the actual USB lib call*
// *                                                                     *
// * Parameters:                                                         *
// *    PSP_SHARED_MEM - pointer to shared memory                        *
// *                                                                     *
// * Return Codes:                                                       *
// *    RC_RESET          - USB diconnection, restart ibmasm             *
// *    RC_INTERRUPTED    - Interrupted by a signal, terminate deamon    *
// *    RC_INVALID_HANDLE - Invalid parm passed by caller                *
// *    RC_SUCCESS        - Successful completion                        *
// ***********************************************************************
int CommandHandler(PSP_SHARED_MEM pSpSharedMem)
{
    int i;
    int rc;
    int node;
    int Handle;
    int CmdLen;
    int BuffLen;
	 int RetryCount;						

    PSP_DEV_NODE pDevNode;

    // Start of PHR_223662
    int NumSignals;

    /* pointer to a signal handler */
    typedef void (*sighandler_t)(int);
    sighandler_t sig;

    // SIGNAL HANDLER  -  cannot catch SIGKILL or SIGSTOP
    int SigArray[] = {SIGINT, SIGTERM, SIGUSR1, SIGUSR2};

    NumSignals = (sizeof(SigArray) / sizeof(SigArray[0]));

    // Enable signal handling
    for (i = 0; i < NumSignals; i++)
    {
        sig = signal(SigArray[i], sig_handler);

        if (sig == SIG_ERR)
        {
           if(debug) syslog(LOG_CRIT, "signal() returned SIG_ERR, unable to register for signal. %i.\n", SigArray[i]);
           return EXIT_FAILURE;
        }
    }
    // End of PHR_223662


    // INITIALIZE GLOBALS
    InterruptedBySignal = FALSE;
    CommandReceived     = FALSE;
    ResetUSB            = FALSE;

    // THIS IS WHERE THE BULK OF THE WORK GOES
    while(TRUE)
    {
    	rc = RC_SUCCESS;

        //if(debug) syslog(LOG_CRIT, "Waiting for a command request from the shared obj.\n");
        if (InterruptedBySignal)
        {
	        InterruptedBySignal = FALSE;
	        rc = RC_INTERRUPTED;
	        break;
        }

        // PHR_225158
    	// ResetUSB IS SET BY THE EventHandler() IF IT DETECTS A CHANGE IN NODE STATUS
    	if (ResetUSB)	
    	{
            if(debug) syslog(LOG_CRIT, "CommandHandler: Received a ResetUSB from EventHandler, returning RC_RESET.\n");
    	    ResetUSB = FALSE;
    	    rc = RC_RESET;
    	    break;
    	}

        // WAIT ON THE SIGNAL (SIGUSR1) THAT INDICATES THERE IS AN
        // OUTSTANDING COMMAND TO BE PROCESSED
        sleep(5);

        while (CommandReceived)
        {
            CommandReceived = FALSE;

            node = pSpSharedMem->CallersNode;

            if ( (node < 0) || (node >= MAX_SP_NODES) )
            {
                if(debug) syslog(LOG_CRIT, "CommandHandler():  detected a invalid node, returning RC_INVALID_HANDLE.\n");
	            rc = RC_INVALID_HANDLE;
                pSpSharedMem->CallerList[Handle].ReturnCode = rc;
                pSpSharedMem->CmdRespReceived = TRUE;
	            break;
            }

            Handle = pSpSharedMem->CallersHandle;

            if ( (Handle <= 0) ||  (Handle > MAX_HANDLES)  )
            {
                if(debug) syslog(LOG_CRIT, "CommandHandler(): detected a invalid handle, returning RC_INVALID_HANDLE.\n");
	            rc = RC_INVALID_HANDLE;
                pSpSharedMem->CallerList[Handle].ReturnCode = rc;
                pSpSharedMem->CmdRespReceived = TRUE;
	            break;
            }

            //if(debug) syslog(LOG_CRIT, "CmdReq received, from PID = %i, node = %i and handle =%i.\n",
            //        getpid(), node, Handle);

            CmdLen = (int)(pSpSharedMem->Buffer)[3];

            for (i = 0; i <= CmdLen; i++)
            {
                pSpSharedMem->CmdBuff[i] = pSpSharedMem->Buffer[8+i];
            }

            // GET A POINTER TO THE CALLERS BUFFER AND THE AMT OF DATA IN THAT BUFFER
            BuffLen = pSpSharedMem->BuffLen;

            // ADD 2 BYTES OF USB BUFFER LENGTH TO THE CALLERS BUFFER LENGTH
            BuffLen += 2;

	         //if(debug) syslog(LOG_CRIT, "Sending command %s to the SP adapter.\n",pSpSharedMem->CmdBuff);
	         pDevNode = &pSpSharedMem->SPDevNode[node];

		      // LOOP INITIALIZER
		      rc = RC_WRITE_ERROR;
           
		      while (rc != RC_SUCCESS)
		      {
		      	rc = SendCmdToSP(pSpSharedMem, pDevNode, pSpSharedMem->Buffer, BuffLen, MAX_RETRYS );
           
		      	if (rc == RC_SUCCESS)
		      	{
                  pSpSharedMem->CallerList[Handle].ReturnCode = rc;
		      		pSpSharedMem->CmdRespReceived = TRUE;
		      		break;
               }
		      	else  // TIMEOUT OR WRITE ERROR
		      	{
                  RetryCount++;
		      		
		      		if((rc == RC_WRITE_ERROR) || (RetryCount >= MAX_RETRYS) || (rc == RC_READ_ERROR))
		      		{
		      			if(debug) syslog(LOG_CRIT, "CommandHandler(): RC_WRITE_ERROR or >= MAX_RETRYS\n");
		      			//pSpSharedMem->Resetting = TRUE;        // PHR_225158
                     pDevNode->State = NODE_DISCONNECTED;     // PHR_225158
		      			pSpSharedMem->CallerList[Handle].ReturnCode = rc;
                     ResetUSB = TRUE;
		      			pSpSharedMem->CmdRespReceived = TRUE;
                     break;
                  }
		      	}
            }

            break;
        }   // end while Command Received
    }       // END WHILE (FOREVER)

    return rc;
}


// ***********************************************************************
// * EventHandler() - This routine polls the adapter for events, if an   *
// *                  event occurred it wakes all processes waiting for  *
// *                  events. An array index is used to keep track  of   *
// *                  the last event written. The registered processes   *
// *                  also use this index to determine if any events were*
// *                  missed between registrations, if an event was      *
// *                  missed it will return the missed event immediately *
// *                                                                     *
// * Parameters:                                                         *
// *    PSP_SHARED_MEM - pointer to shared memory                        *
// *                                                                     *
// * Return Codes:                                                       *
// *    RC_RESET       - USB diconnection, restart ibmasm                *
// *    RC_INTERRUPTED - Interrupted by a signal, terminate deamon       *
// *    RC_SUCCESS     - Successful completion                           *
// ***********************************************************************
int EventHandler(PSP_SHARED_MEM pSpSharedMem)
{
    int node;
    int rc;
    int EventOccurred;
    int LockFileDesc;
    int NodesPolled;               // PHR_225158
    int NodesEnabled;              // PHR_225158

    int SpCount     = 0;
    int ScanCounter = 0;

    // Start of PHR_223662
    int NumSignals;

    /* pointer to a signal handler */
    typedef void (*sighandler_t)(int);
    sighandler_t sig;

    // SIGNAL HANDLER  -  cannot catch SIGKILL or SIGSTOP
    int i;
    int SigArray[] = {SIGINT, SIGTERM};

    NumSignals = (sizeof(SigArray) / sizeof(SigArray[0]));

    // Enable signal handling
    for (i = 0; i < NumSignals; i++)
    {
        sig = signal(SigArray[i], sig_handler);

        if (sig == SIG_ERR)
        {
           if(debug) syslog(LOG_CRIT, "signal() returned SIG_ERR, unable to register for signal. %i.\n", SigArray[i]);
           return EXIT_FAILURE;
        }
    }
    // End of PHR_223662

    pSpSharedMem->ChildsPid = getpid();

    rc = RC_SUCCESS;
    ScanCounter = 0;

    while (!pSpSharedMem->Resetting && !InterruptedBySignal)         // PHR_225158
    {
        // Start of PHR_220505
        // SCAN THE USB DEVICES ONCE EVERY 10 ITERATIONS (20 SECONDS) TO SEE IF ANY
        // USB SP DEVICES WERE ADDED. IF THERE WAS, RE-ENUMERATE AND RE-INIT ALL
        // DEVICES. ALSO USE THIS COUNTER TO ENABLE SHUTDOWN EVENTS ON THE SP.
        if(ScanCounter < 10)
        {
            ScanCounter++;
        }
        else
        {
            if (pSpSharedMem->Resetting || InterruptedBySignal)
            {
                break;
            }

            ScanCounter = 0;

            SpCount = 0;

            LockFileDesc = LockSharedMemory();

            if (LockFileDesc != -1)
            {
                //if(debug) syslog(LOG_CRIT, "EventHandler() locked shared memory for counting SP's. \n");

                SpCount = CountSPs(pSpSharedMem);

                if (SpCount > pSpSharedMem->SpCount)
                {
                   if(debug) syslog(LOG_CRIT, "EventHandler() detected that an RSA II device has been added.\n");
                   UnlockSharedMemory(LockFileDesc);
                   //if(debug) syslog(LOG_CRIT, "EventHandler() unlocked shared memory to count SP's. \n");
                   return RC_RESET;
                }

                UnlockSharedMemory(LockFileDesc);
                //if(debug) syslog(LOG_CRIT, "EventHandler() unlocked shared memory for counting SP's. \n");
            }
            else
            {
            	if(debug) syslog(LOG_CRIT, "EventHandler() could not lock shared memory to count SP's. \n");
            }

            NodesEnabled = 0;

            // ENABLE SHUTDOWN EVENTS FOR ALL NODES
            for (node = 0; node < MAX_SP_NODES; node++)
            {
                if (pSpSharedMem->Resetting || InterruptedBySignal)
                {
                    break;
                }

                if (pSpSharedMem->SPDevNode[node].State != NODE_AVAILABLE)
                {
                    NodesEnabled++;

                    if (pSpSharedMem->SPDevNode[node].State == NODE_OPENED)
                    {
                       LockFileDesc = LockSharedMemory();

                       if (LockFileDesc != -1)
                       {
                            //if(debug) syslog(LOG_CRIT, "EventHandler() locked shared memory to enable shutdown events. \n");

                            rc = EnableShutdownEvent(pSpSharedMem, node);

                            UnlockSharedMemory(LockFileDesc);
                            //if(debug) syslog(LOG_CRIT, "EventHandler() unlocked shared memory for enabling shutdown events. \n");
                       }
                       else
                       {
                       	if(debug) syslog(LOG_CRIT, "EventHandler() could not lock shared memory to enable shutdown events. \n");
                       }
                    }

                    if (NodesEnabled >= pSpSharedMem->SpCount)
                    {
                        break;
                    }
                    else
                    {
                        // LET OTHER TASKS GET ACCESS TO SHARED MEMORY IN BETWEEN ENABLES
                        sleep(1);
                    }
                }
            }
            // End of PHR_220505
        }

        if (pSpSharedMem->Resetting || InterruptedBySignal)
        {
            rc = RC_RESET;
            break;
        }

        rc = RC_SUCCESS;
        NodesPolled = 0;                                         // PHR_225158

        for (node = 0; node < MAX_SP_NODES; node++)              // PHR_225158
	     {
           if (InterruptedBySignal || pSpSharedMem->Resetting)  // PHR_225158
           {
               rc = RC_RESET;
               break;
           }

           EventOccurred = FALSE;

	        if(pSpSharedMem->SPDevNode[node].State != NODE_AVAILABLE)
	        {
                NodesPolled++;
           }                            // PHR_225158

           // ONLY POLL IF THE NODE IS OPENED
           if(pSpSharedMem->SPDevNode[node].State == NODE_OPENED)      // PHR_225158
           {                                                           // PHR_225158
              LockFileDesc = LockSharedMemory();

	           if (LockFileDesc != -1)
	           {
                  //if(debug) syslog(LOG_CRIT, "EventHandler() Locked shared memory to poll for events.\n");

                  EventOccurred = PollForEvents(pSpSharedMem, node);

	               if (EventOccurred)
	               {
	                   pSpSharedMem->SPDevNode[node].EventIndex += 1;

	                   if (pSpSharedMem->SPDevNode[node].EventIndex == EVENT_QUEUE_DEPTH)
	                   {
	                       pSpSharedMem->SPDevNode[node].EventIndex = 0;
	                   }

	                   //if(debug) syslog(LOG_CRIT, "Event occurred, waking up waiting tasks on node %i.\n",node);
	                   EventOccurred = FALSE;
	                   WakeEventTasks(pSpSharedMem, node);
                      UnlockSharedMemory(LockFileDesc);

                      // LET OTHER TASKS GET ACCESS TO SHARED MEMORY
                      sleep(1);
                      //if(debug) syslog(LOG_CRIT, "EventHandler() unlocked shared memory for polling for events.\n");
                  }
                  else
                  {
                      UnlockSharedMemory(LockFileDesc);
                      //if(debug) syslog(LOG_CRIT, "EventHandler() unlocked shared memory for polling for events.\n");
                  }
              }
	           else
	           {
	           	   if(debug) syslog(LOG_CRIT, "EventHandler could not get shared memory for PollForEvents().\n");
	           }

              // DDH:  moved this check inside "if...OPENED" so only flag reset if state changes from OPENED
              // WHEN PollForEvents() RETURNS THE NODE STATE COULD HAVE CHANGED IF IT
              // COULD NOT COMMUNICATE WITH THE NODE OR THE NODE RECEIVED A RESET EVENT
              if(pSpSharedMem->SPDevNode[node].State != NODE_OPENED)
              {
                  if(debug) syslog(LOG_CRIT, "EventHandler: detected lost USB connection for node %x, state = %x, returning RC_RESET.\n", node, pSpSharedMem->SPDevNode[node].State);
                  rc = RC_RESET;
                  break;
              }
   
           }   // END IF NODE_OPENED                         // PHR_225158

           // LET OTHER TASKS GET ACCESS TO SHARED MEMORY IN BETWEEN POLLS
           if (NodesPolled >= pSpSharedMem->SpCount)
           {
               break;
           }
           else
           {
               sleep(1);
           }
	    }       // END FOR MAX_SP_NODES

       // PHR_225158
       if (rc != RC_SUCCESS)
       {
           break;    // EXIT WHILE
       }

		 // WAIT BEFORE PROCESSING NEXT POLL
       sleep(2);
    }   // END WHILE !InterruptedBySignal

    if (InterruptedBySignal)
    {
       if(debug) syslog(LOG_CRIT, "Eventhandler():  interruptd by a signal\n");
       rc = RC_INTERRUPTED;
    }

    return rc;
}



// ***********************************************************************
// * InitSharedMem() - This routine simply initializes the shared memory *
// *                   block.                                            *
// *                                                                     *
// * Parameters:                                                         *
// *    PSP_SHARED_MEM - pointer to shared memory                        *
// *                                                                     *
// * Return Codes:                                                       *
// ***********************************************************************
void InitSharedMem(PSP_SHARED_MEM pSpSharedMem)
{
   int i;

   // INITIALIZE THE MEMORY JUST ALLOCATED
   memset(pSpSharedMem, 0, sizeof(SP_SHARED_MEM));

   pSpSharedMem->InitialLoad   =  TRUE;
   pSpSharedMem->InitialLog    =  TRUE;
   pSpSharedMem->CallersHandle = -1;
   pSpSharedMem->CallersNode   = -1;
   pSpSharedMem->BuffLen       =  0;

   pSpSharedMem->IbmasmsPid    = getpid();
   pSpSharedMem->ChildsPid     = -1;


   pSpSharedMem->Resetting       = FALSE;
   pSpSharedMem->CmdRespReceived = FALSE;

   // INITIALIZE THE SP's HANDLE LIST, CALLER LIST AND EVENT LIST
   for ( i = 1; i <= MAX_HANDLES; i++)
   {
       pSpSharedMem->HandleList[i].Free   = TRUE;
       pSpSharedMem->HandleList[i].Handle = i;

       pSpSharedMem->CallerList[i].Pid        = -1;
       pSpSharedMem->CallerList[i].Node       = -1;
       pSpSharedMem->CallerList[i].BufferSize =  0;
       pSpSharedMem->CallerList[i].ReturnCode = -1;
       pSpSharedMem->CallerList[i].RegForRH   = FALSE;

       pSpSharedMem->CallerList[i].EventRegInfo.Pid         = -1;
       pSpSharedMem->CallerList[i].EventRegInfo.Handle      = -1;
       pSpSharedMem->CallerList[i].EventRegInfo.EventIndex  = -1;
       pSpSharedMem->CallerList[i].EventRegInfo.Node        = -1;
       pSpSharedMem->CallerList[i].EventRegInfo.Waiting     = FALSE;
       pSpSharedMem->CallerList[i].EventRegInfo.Initialized = FALSE;
       pSpSharedMem->CallerList[i].EventRegInfo.Deregister  = FALSE;
   }


   // INITIALIZE THE SP's DEVICE NODE LIST
   for (i = 0; i < MAX_SP_NODES; i++)
   {
       pSpSharedMem->SPDevNode[i].State        = NODE_AVAILABLE;
       pSpSharedMem->SPDevNode[i].OutEndPoint  = 0;
       pSpSharedMem->SPDevNode[i].InEndPoint   = 0;
       pSpSharedMem->SPDevNode[i].BufferSize   = 0;
       pSpSharedMem->SPDevNode[i].ErrorCount   = 0;
       pSpSharedMem->SPDevNode[i].EventIndex   = 0;
       pSpSharedMem->SPDevNode[i].pUsbHandle   = (void *)INVALID_USB_HANDLE;
   }

   return;
}


// ***********************************************************************
// * ScanRespPkt() - This routine parses a command response packet that  *
// *                 was received from the SP. It validates the command  *
// *                 field, ensuring that it syncs with the command that *
// *                 was sent to the adapter, verifies that this packet  *
// *                 is a command response, then calculates and returns  *
// *                 the size of the response packet.                    *
// *                                                                     *
// * Parameters:                                                         *
// *    PSP_SHARED_MEM - pointer to shared memory                        *
// *    UCHAR *        - pointer to the command response buffer          *
// *                                                                     *
// * Returns:                                                            *
// *    The amout of data returned in the response buffer or -1 if the   *
// *    command response was deemed invalid                              *
// *                                                                     *
// ***********************************************************************
int ScanRespPkt(PSP_SHARED_MEM pSpSharedMem, unsigned char *pBuffer)
{
    unsigned char *TempBuff;
    int            BuffLen;
    int            BytesToRead;
    int            i;

    TempBuff = pBuffer;

    BuffLen =  CMD_BLK_LEN  +          // 6 bytes of cmd header
               TempBuff[3]  +          // cmd length
               TempBuff[4]  +          // data length LSB
              (TempBuff[5]  * 256);    // data length MSB

    // CHECK FOR A VALID COMMAND RESPONSE
    if (TempBuff[2] != SP_CMDRESP)
    {
	    //if(debug) syslog(LOG_CRIT, "ScanResp() found an invalid command type in the command response.\n");
	    return -1;
    }

    // CHECK FOR VALID COMMAND RETURNED
    for (i = 0; i < TempBuff[3]; i++)
    {
        if (TempBuff[8+i] != pSpSharedMem->CmdBuff[i])
        {
	        //if(debug) syslog(LOG_CRIT, "ScanRespPkt() found an invalid command in the command response.\n");
            return -1;
        }
    }

    BytesToRead = BuffLen + CHK_SUM_LEN + USB_PKT_LEN;

    if (BytesToRead >= MAX_BUFF_SIZE)
    {
       //if(debug) syslog(LOG_CRIT, "ScanRespPkt() found an invalid response length in the command response.\n");
       return -1;
    }

    pSpSharedMem->BuffLen = BuffLen;
    return BytesToRead;
}




// ***********************************************************************
// * ScanEventPkt() - This routine parses an event response packet that  *
// *                 was received from the SP. It validates the command  *
// *                 field, ensuring that it syncs with the command that *
// *                 was sent to the adapter, in this case it is always  *
// *                 a poll for events (5.2)  verifies that this packet  *
// *                 is a command response, then calculates and returns  *
// *                 the size of the response packet.                    *
// *                                                                     *
// *                 NOTE:  This routine also checks byte 18 of the resp *
// *                        buffer to see if this is a request to halt   *
// *                        the OS. This is done in some cases prior to  *
// *                        the adapter rebooting the system so the OS   *
// *                        can be brought down gracefully.              *
// *                                                                     *
// * Parameters:                                                         *
// *    UCHAR *        - pointer to the event polls response buffer      *
// *                                                                     *
// * Returns:                                                            *
// *    The amout of data returned in the response buffer or -1 if the   *
// *    events command response was deemed invalid                       *
// *                                                                     *
// ***********************************************************************
int ScanEventPkt(PSP_SHARED_MEM pSpSharedMem, int node, unsigned char *EventBuff)  // PHR_225158
{
    unsigned char *TempBuff;
    int            BuffLen;
    int            BytesToRead;

    TempBuff = EventBuff;

    BuffLen =  CMD_BLK_LEN  +
               TempBuff[3]  +          // cmd length
               TempBuff[4]  +          // data length LSB
              (TempBuff[5]  * 256);    // data length MSB

    // CHECK FOR VALID EVENT RETURNED
    if ( (TempBuff[2] != SP_CMDRESP) ||
         (TempBuff[8] != 5)          ||
         (TempBuff[9] != 2)          )
    {
        //if(debug) syslog(LOG_CRIT, "ScanRespPkt() found invalid cmd data in resp pkt, returning failure.\n");
        return -1;
    }

    if (TempBuff[18] == 4)
    {
        if(debug) syslog(LOG_CRIT, "ScanEventPkt() Received shutdown notification from SP adapter - Issueing system Halt\n");
        system("/sbin/halt");
    }

    // Start of PHR_225158
    if (TempBuff[18] == 16)
    {
        //if(debug) syslog(LOG_CRIT, "ScanEventPkt() detected a Reset Event from the SP adapter. ibmasm will re-initialize.\n");
        pSpSharedMem->SPDevNode[node].State = NODE_RESETTING;
    }
    // End of PHR_225158

    BytesToRead = BuffLen + CHK_SUM_LEN + USB_PKT_LEN;

    if ( BytesToRead > EVENT_BUFF_SIZE )
    {
       //if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_bulk_read returned an invalid Event Response Length = %i.\n",BuffLen);    // PHR_LINK
       return -1;
    }

    return BytesToRead;
}







// ***********************************************************************
// * SendInitCmds() - This routine initializes a command buffer with the *
// *                  two initilization commands and sends those commands*
// *                  to the SP vis SendCmdToSP(). It will retry a max of*
// *                  3 times for each of these commands.                *
// *                                                                     *
// * Parameters:                                                         *
// *    PSP_SHARED_MEM - pointer to shared memory                        *
// *    PSP_DEV_NODE   - pointer to the device node to send the cmd to   *
// *                                                                     *
// * Returns:                                                            *
// *    RC_SUCCESS       - successful completion                         *
// *    RC_SYSTEM_ERROR  - Command read or write failed                  *
// *                                                                     *
// ***********************************************************************
int SendInitCmds(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode)
{
   unsigned char InitBuff[1024];

   int RetryCount;
   int rc;
   int i;

   //*****************************************************************
   // SEND THE 4.3.5.A (DRIVER VPD) TO THE SP                        *
   //*****************************************************************
   rc = RC_READ_ERROR;  // INITIALIZER
   RetryCount = 0;

   while (rc != RC_SUCCESS)
   {
       memset(InitBuff, 0, 1024 );

       // Fill the buffer with spaces
       for (i = 0; i < 27; i++ )
       {
          InitBuff[i] = ' ';
       }

       InitBuff[0] = (unsigned char)((11 & 0xFF00)>>8 ); // USB pkt length (LSB)
       InitBuff[1] = (unsigned char)( 11 & 0xFF);        // USB pkt length (MSB)
       InitBuff[2]  = SP_WRITE;                          // Command Type
       InitBuff[3]  = 4;                                 // Command Length
       InitBuff[4]  = 16;                                // Data Length (LSB)
       InitBuff[5]  = 0;                                 // Data Length (MSB)
       InitBuff[6]  = 0;                                 // Status
       InitBuff[7]  = 0;                                 // Reserved
       InitBuff[8]  = 4;                                 // Command Byte 0
       InitBuff[9]  = 3;                                 // Command Byte 1
       InitBuff[10] = 5;                                 // Command Byte 2
       InitBuff[11] = 0x0a;                              // Command Byte 3
       InitBuff[12] = 'L';
       InitBuff[13] = 'I';
       InitBuff[14] = 'N';
       InitBuff[15] = '3';
       InitBuff[16] = '2';
       InitBuff[17] = ' ';
       InitBuff[18] = '1';      // for build ibmusbasm-1.46-2.i386.rpm
       InitBuff[19] = '.';
       InitBuff[20] = '4';
       InitBuff[21] = '6';

       InitBuff[27]   = '\0';
       InitBuff[28]   = usb_cksum(InitBuff, 26);

       #ifdef IA64
         InitBuff[15] = '6';
         InitBuff[16] = '4';
       #endif

       #ifdef __x86_64__
         InitBuff[15] = '6';
         InitBuff[16] = '4';
       #endif

       pSpSharedMem->CmdBuff[0] = 4;
       pSpSharedMem->CmdBuff[1] = 3;
       pSpSharedMem->CmdBuff[2] = 5;
       pSpSharedMem->CmdBuff[3] = 0x0a;

       //if(debug) syslog(LOG_CRIT, "SendInitCmd(), sending Init command 4.3.5.A to the SP.\n");

       // DDH Handle retry differently on first cmd to SP because I have seen that cmd disappear frequently on some systems
       rc = SendCmdToSP(pSpSharedMem, pDevNode, InitBuff, 28, RetryCount+1);

       if (rc != RC_SUCCESS)
       {
           RetryCount++;

           if (RetryCount >= MAX_RETRYS+2)
           {
               //if(debug) syslog(LOG_CRIT, "SendInitCmd(), Init command 4.3.5.A failed, returning Failure.\n");
               return RC_SYSTEM_ERROR;
           }
       }
   }

   //*****************************************************************
   // SEND THE 4.3.6 INIT COMMAND                                    *
   //*****************************************************************
   RetryCount = 0;
   rc = RC_READ_ERROR;  // INITIALIZER

   while (rc != RC_SUCCESS)
   {
       memset(InitBuff, 0, 1024 );

       InitBuff[0] = (unsigned char)((11 & 0xFF00)>>8 ); // USB pkt length (LSB)
       InitBuff[1] = (unsigned char)( 11 & 0xFF);        // USB pkt length (MSB)

       InitBuff[2]  = SP_WRITE;                          // Command Type
       InitBuff[3]  = 3;                                 // Command Length
       InitBuff[4]  = 1;                                 // Data Length  LSB
       InitBuff[5]  = 0;                                 // Data Length  MSB
       InitBuff[6]  = 0;                                 // Status
       InitBuff[7]  = 0;                                 // Reserved
       InitBuff[8]  = 4;                                 // Command Byte 0
       InitBuff[9]  = 3;                                 // Command Byte 1
       InitBuff[10] = 6;                                 // Command Byte 2
       InitBuff[11] = 5;                                 // Data Byte 0

       InitBuff[12] = usb_cksum(InitBuff, 10);           // USB check sum

       pSpSharedMem->CmdBuff[0] = 4;
       pSpSharedMem->CmdBuff[1] = 3;
       pSpSharedMem->CmdBuff[2] = 6;

       //if(debug) syslog(LOG_CRIT, "SendInitCmd(), sending Init command 4.3.6 to the SP.\n");

       rc = SendCmdToSP(pSpSharedMem, pDevNode, InitBuff, 13, MAX_RETRYS);

       if (RetryCount >= MAX_RETRYS)
       {
           //if(debug) syslog(LOG_CRIT, "SendInitCmd(), Init command 4.3.6 failed, returning Failure.\n");
           return RC_SYSTEM_ERROR;
       }
   }

   return RC_SUCCESS;
}




// ***********************************************************************
// * CreateLockFile() - This routine creates /var/run/ibmasm.pid a lock  *
// *                    file that ensures that only one instance of      *
// *                    ibmasm can be running at a time.                 *
// *                                                                     *
// * Parameters:                                                         *
// *                                                                     *
// * Returns:                                                            *
// *    RC_SUCCESS       - Successful completion                         *
// *    EXIT_FAILURE     - Could not open the lock file                  *
// *                                                                     *
// ***********************************************************************
int CreateLockFile(void)
{
    int  rc;
    int  len;
    int  i;
    int  CreateFile;

    char popenStr[70] = "ps -p 00000 --no-headers";
    char IbmAsmPid[10];
    char ChkData[10];

    FILE *pLockFile;
    FILE *pChkLock;

    CreateFile = TRUE;

    // OPEN THE IBMASM LOCK FILE FOR READ ONLY, IF IT OPENS SUCCESSFULLY
    // THAT MEANS THAT IT EXISTS SO IBMASM MAY ALREADY BE ACTIVE, IT MAY
    // ALSO MEAN THAT IT WAS CREATED AND NOT REMOVED BECAUSE OF AN ABEND
    // SO CHECK FOR THAT CONDITION AND IF TRUE DELETE AND RECREATE THE FILE
    pLockFile = fopen( ASM_PID_FILE, "r");

    while(pLockFile)
    {
        // INITIALIZE THE STRINGS
        for (i = 0; i < 10; i++)
        {
            IbmAsmPid[i] = 0x0;
            ChkData[i]   = 0x0;
        }

        // GET THE ORIGINATORS PID FROM IBMASM.PID
        fscanf(pLockFile, "%s", IbmAsmPid);
        fclose(pLockFile);

        len = strlen(IbmAsmPid);

        // IF THE FILE IS EMPTY REMOVE THE FILE AND RECREATE IT
        if (IbmAsmPid[0] == 0x0)
        {
           unlink(ASM_PID_FILE);
           CreateFile = TRUE;
           break;
        }

        // FILL THE PID INTO THE PS -P SYSTEM COMMAND
        for ( i = 0; i < len; i++)
        {
            popenStr[6+i] = IbmAsmPid[i];

        }

        // PAD WITH BLANKS
        for ( ; i < 6; i++)
        {
            popenStr[6+i] = 0x20;
        }

        // ISSUE THE PS -P COMMAND, THIS WILL CREATE A FILE CONTIANING
        // THE ORIGINATORS PID IF IT EXISTS, OTHERWISE IT CREATES AN
        // EMPTY FILE
        //system(SysCmdStr);

        //if(debug) syslog(LOG_CRIT, "ps_string = %s \n", SysCmdStr);

        pChkLock = popen(popenStr, "r");
        fscanf(pChkLock, "%s", ChkData);
        pclose(pChkLock);

        //if(debug) syslog(LOG_CRIT, "ChkData = %s \n", ChkData);

        // IF THE FILE IS EMPTY, THE ORIGINATOR IS GONE SO KILL THE
        // FILE AND CREATE A NEW ONE, ELSE RETURN FAILURE BECAUSE
        // IBMASM IS ALRADY LOADED
        if (ChkData[0] == 0x0)
        {
            //if(debug) syslog(LOG_CRIT, "Lock file existed but creator had terminated, continuing load. \n");
            unlink(ASM_PID_FILE);
            CreateFile = TRUE;
            break;
        }
        else
        {
            //if(debug) syslog(LOG_CRIT, "Lock file already exists, returning EXIT_FAILURE. \n");
            CreateFile = FALSE;
            break;
        }
    }

    if (CreateFile)
    {
        pLockFile = fopen( ASM_PID_FILE, "w");

        if ( pLockFile )
        {
            fprintf(pLockFile,"%d\n", getpid() );
            fclose( pLockFile );
            rc = RC_SUCCESS;
        }
        else
        {
            rc = EXIT_FAILURE;
        }
    }
    else
    {
        rc = EXIT_FAILURE;
    }

    return rc;
}







// ***********************************************************************
// * usb_cksum() - This routine calculates and returns a check sum based *
// *               on the contents of the buffer being sent to the SP.   *
// *                                                                     *
// * Parameters:                                                         *
// *    unsigned char *  - the buffer used to calculate the check sum    *
// *    int              - the amount of data in the buffer              *
// *                                                                     *
// * Returns:                                                            *
// *    unsigned char - The calculated check sum value                   *
// *                                                                     *
// ***********************************************************************
unsigned char  usb_cksum( unsigned char* buf, int size)
{
   unsigned char sum;

   sum = 0;

   while (size > 0)
   {
      sum += *(unsigned short*)buf;        // add the next word to "sum"
      buf += 1;                            // point to next word in slim packet
      size -= 1;                           // dec "size" by 2 bytes
   }

   sum = 1 + ~sum;
   return sum;
}



// ***********************************************************************
// * PollForEvents() - This routine issues a 5.2 Poll for event request  *
// *                   to the adapter. If an event is present the data   *
// *                   size in the response buffer will be non-zero      *
// *                   therefore it will return true.                    *
// *                                                                     *
// * Parameters:                                                         *
// *    PSP_SHARED_MEM - pointer to shared memory                        *
// *    int            - the device node to send the cmd to              *
// *                                                                     *
// * Returns:                                                            *
// *    TRUE   -   Indication that there is an outstanding event present *
// *    FALSE  -   There is no event present                             *
// *                                                                     *
// ***********************************************************************
int PollForEvents(PSP_SHARED_MEM pSpSharedMem, int node)
{
   int rc;
	int ErrorCount;
   int FirstPkt;
   int BytesToRead;
   int BytesRead;
   int BytesWritten;
   int TimeOut = 0;

   unsigned char *TempPtr;
   unsigned char  EventBuff[EVENT_BUFF_SIZE];

   memset(EventBuff, 0, EVENT_BUFF_SIZE);

	ErrorCount   = 0;
	BytesWritten = 0;
	TimeOut      = 0;

   while (ErrorCount < 10)		
	{
        EventBuff[0] = (unsigned char)((9 & 0xFF00)>>8 ); // USB pkt length (LSB)
        EventBuff[1] = (unsigned char)( 9 & 0xFF);        // USB pkt length (MSB)

        EventBuff[2]  = SP_READ;                  // Command Type
        EventBuff[3]  = 2;                        // Command Length
        EventBuff[4]  = 0;                        // Data Length  LSB
        EventBuff[5]  = 0;                        // Data Length  MSB
        EventBuff[6]  = 0;                        // Status
        EventBuff[7]  = 0;                        // Reserved
        EventBuff[8]  = 5;                        // Command Byte 0
        EventBuff[9]  = 2;                        // Command Byte 1
        EventBuff[10] = usb_cksum(EventBuff, 8);  // USB check sum 6;                        // Command Byte 2
        EventBuff[11] = 0x0;                      // Data Byte 0

	 	 if ( TimeOut >= CMD_TIMEOUT)
		 {
		    //if(debug) syslog(LOG_CRIT, "PollForEvents cannot communicate with the SP, returning with reset flag set.\n");
		    //pSpSharedMem->Resetting = TRUE;		                         // PHR_225158
          pSpSharedMem->SPDevNode[node].State = NODE_DISCONNECTED;         // PHR_225158
			 return FALSE;
		 }

       // SEND EVENT POLL TO THE ADAPTER
       while(BytesWritten != MAX_WRITE_LEN)
       {
            BytesWritten = pSpSharedMem->sp_usb_bulk_write( pSpSharedMem->SPDevNode[node].pUsbHandle, 
                                                            pSpSharedMem->SPDevNode[node].OutEndPoint,
                                                            EventBuff, 
                                                            MAX_WRITE_LEN, 
                                                            3000);  // PHR_LINK

	        // IF A WRITE FAILURE OCCURS START OVER
	        if (BytesWritten != MAX_WRITE_LEN)
	        {
                TimeOut += 3;

                if (TimeOut > CMD_TIMEOUT)
	            {
					//if(debug) syslog(LOG_CRIT, "PollForEvents cannot communicate with the SP, returning with reset flag set.\n");
                    //pSpSharedMem->Resetting = TRUE;		                         // PHR_225158
                    pSpSharedMem->SPDevNode[node].State = NODE_DISCONNECTED;         // PHR_225158
	                return FALSE;
	            }
            }
       }

       // WAIT FOR THE COMMAND RESPONSE FROM THE ADAPTER
       FirstPkt    = TRUE;
       TempPtr     = EventBuff;
       BytesToRead = 1;   // Initializer

       while (BytesToRead > 0)
       {
            //if(debug) syslog(LOG_CRIT, "EVENT_HNDLR: send of event poll succeded, waiting for CmdResp.\n");

            BytesRead = pSpSharedMem->sp_usb_bulk_read( pSpSharedMem->SPDevNode[node].pUsbHandle, 
                                                        pSpSharedMem->SPDevNode[node].InEndPoint,
                                                        TempPtr, 
                                                        MAX_READ_LEN, 
                                                        5000);   // PHR_LINK

			// TRY TO READ THE COMMAND RESPONSE FOR 15 SECONDS (3 TIMES)
			// IF UNSUCESSFUL, RETRY FROM THE START (WRITE AGAIN)
            while (BytesRead != MAX_READ_LEN)
            {
                ErrorCount += 1;
                TimeOut    += 5;

                if ( TimeOut >= CMD_TIMEOUT)
				{
					//if(debug) syslog(LOG_CRIT, "PollForEvents cannot communicate with the SP, returning with reset flag set.\n");
               //pSpSharedMem->Resetting = TRUE;		                         // PHR_225158
               pSpSharedMem->SPDevNode[node].State = NODE_DISCONNECTED;         // PHR_225158
					return FALSE;
				}

            if (ErrorCount >= 3 )
            {
					break;
            }

            BytesRead = pSpSharedMem->sp_usb_bulk_read( pSpSharedMem->SPDevNode[node].pUsbHandle, 
                                                            pSpSharedMem->SPDevNode[node].InEndPoint, 
                                                            TempPtr, 
                                                            MAX_READ_LEN, 
                                                            5000);   // PHR_LINK
            }

            if (ErrorCount >= 3 )
		      {
			      break;
			   }

            // CALCULATE THE LENGTH OF THE COMMAND DATA TO BE RECIEVED,
            // THIS IS BASED ON THE CONTENTS OF THE FIRST CMD RESP PKT
            if (FirstPkt)
            {
                FirstPkt = FALSE;
                BytesToRead = ScanEventPkt(pSpSharedMem, node, TempPtr);

                if (BytesToRead == -1)
                {
                    BytesRead = MAX_READ_LEN;

                    // READ UNTIL A USB READ ERROR OCCURS, THIS IS
                    // DONE SO WE CLEAR ALL UNWANTED DATA IN THE BUFFER
                    while(BytesRead == MAX_READ_LEN)
                    {
                        BytesRead = pSpSharedMem->sp_usb_bulk_read( pSpSharedMem->SPDevNode[node].pUsbHandle, 
                                                                    pSpSharedMem->SPDevNode[node].InEndPoint,
                                                                    TempPtr, 
                                                                    MAX_READ_LEN, 
                                                                    5000);   // PHR_LINK
                    }

                    //if(debug) syslog(LOG_CRIT, "Read error occurred while polling for events on node %i, Incrementing error count.\n", node);
                    return FALSE;
                }
            }

            ErrorCount = 0;
            BytesToRead -= BytesRead;
            TempPtr += BytesRead;
       }   // end while BytesToRead

		 if (ErrorCount == 0)
	    {
		    break;
		 }
    }

    //if(debug) syslog(LOG_CRIT, "EVENT_HNDLR: event poll CmdResp returned successfully.\n");

    // ANY DATA RETURNED IN THE COMMAND RESPONSE INDICATES THAT AN EVENT OCCURRED
    // COPY IT INTO THE EVENT QUEUE AND RETURN TRUE.
    if (EventBuff[4] || EventBuff[5])
    {
        //int i;
        int BuffLen;
        int EventQueueIndex;

        // GET THE LENGTH OF THE EVENT DATA
        BuffLen = 6    +        // Response Header Length
		  EventBuff[1]   +        // cmd length
		  EventBuff[4]   +        // data length LSB
        (EventBuff[5]  * 256);  // data length MSB

        // MAKE SURE IT FITS INTO OUR EVENT BUFFER
        if (BuffLen >= EVENT_BUFF_SIZE)
        {
            //if(debug) syslog(LOG_CRIT, "Event contains more data than buffer can hold, returning RC_INVALID_BUFFER.\n");
            BuffLen = EVENT_BUFF_SIZE - 1;
            //pSpShaaredMem->ReturnCode = RC_INVALID_BUFFER;
        }

        // SET THE TEMP POINTER TO POINT TO THE EVENT RECEIVED (NEEDS TO BE SET BEYOND
        // THE 10 BYTES OF COMMAND RESPONSE DATA RECEIVED FROM THE SP).
        TempPtr  = EventBuff;
        TempPtr += 10;

        // Not sure about this but ...
    	  // Set the command type to SP_EVENT, it may come to ibmasm as 0 from the SP
	     TempPtr[0] = SP_EVENT;

        EventQueueIndex =  pSpSharedMem->SPDevNode[node].EventIndex;

        // CHECK FOR A WRAPPED QUEUE, RESET THE INDEXER IF IT'S AT THE END OF THE QUEUE
        if ( EventQueueIndex == EVENT_QUEUE_DEPTH )
        {
            pSpSharedMem->SPDevNode[node].EventIndex = 0;
            EventQueueIndex = 0;
        }

        // CLEAN OUT THE OLD EVENT, COPY THE NEW EVENT INTO THE QUEUE AND BUMP THE INDEXER
        memset(pSpSharedMem->SPDevNode[node].EventQueue.Event[EventQueueIndex], 0, EVENT_BUFF_SIZE);
        memcpy(pSpSharedMem->SPDevNode[node].EventQueue.Event[EventQueueIndex], TempPtr, BuffLen);

        return TRUE;
    }
    else
    {
        return FALSE;
    }
}














// ***********************************************************************
// * WakeEventTasks() - This routine parses the caller list for processes*
// *                    registered for events. If a waiting process is   *
// *                    found wake it and wait a bit before waking the   *
// *                    nex process thats waiting.                       *
// *                                                                     *
// * Parameters:                                                         *
// *    PSP_SHARED_MEM - pointer to shared memory                        *
// *    int            - the device node to wake                         *
// *                                                                     *
// * Returns:                                                            *
// *                                                                     *
// ***********************************************************************
void WakeEventTasks(PSP_SHARED_MEM pSpSharedMem, int node)
{
    int i;

    for (i = 0; i < MAX_HANDLES; i++)
    {
        if ((pSpSharedMem->CallerList[i].EventRegInfo.Waiting)      &&
            (pSpSharedMem->CallerList[i].EventRegInfo.Node == node) )
        {
           // MAY BE ABLE TO USE A PIPE HERE
           pSpSharedMem->CallerList[i].EventRegInfo.EventReceived = TRUE;

           //if(debug) syslog(LOG_CRIT, "Waking up PID %i.\n", pSpSharedMem->CallerList[i].EventRegInfo.Pid);
           //kill( pSpSharedMem->CallerList[i].EventRegInfo.Pid, SIGUSR1 );
           //usleep(8);
        }
    }
}




//**************************************************************************************
// USE THE FOLLOWING LOCK FILE METHOD TO MAKE SURE THAT IBMASM IS NOT ALREADY LOADED   *
// TRY TO OPEN THE /tmp/ibmasm.pid FILE WITH READ ACCESS. IF IT OPENS SUCCESSFULLY     *
// IBMASM IS ALREADY LOAEDED SO CLOSE THE FILE AND ABORT. OTHERWISE IT WAS NOT CREATED *
// SO OPEN THE FILE WITH WRITE ACCESS, STORE THE PID AND CONTINUE                      *
//**************************************************************************************
int LockSharedMemory(void)
{
    int FileDesc;
    int TimeOut = 0;

    while(TimeOut < CMD_TIMEOUT)
    {
	    FileDesc = -1;

	    FileDesc = open(MEM_LOCK_FILE, O_CREAT | O_EXCL);

        if (FileDesc == -1)
        {
	        TimeOut += 1;
		
           if (TimeOut >= CMD_TIMEOUT)
           {
                //if(debug) syslog(LOG_CRIT, "open of MEM_LOCK_FILE failed with a timeout.\n");
	            return -1;
	        }

	        sleep(1);
	    }
	    else
	    {
	        break;
	    }
    }

    //if(debug) syslog(LOG_CRIT, "locking file descriptor %i.\n",FileDesc);

    return FileDesc;
}


void UnlockSharedMemory(int FileDesc)
{
    //if(debug) syslog(LOG_CRIT, "unlocking file descriptor %i.\n",FileDesc);
    close(FileDesc);
    unlink(MEM_LOCK_FILE);
    //if(debug) syslog(LOG_CRIT, "Shared memory unlocked.\n");
    return;
}


void DeleteLockFile(void)
{
    unlink(ASM_PID_FILE);
    return;
}


int GetMacAddr(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode)
{
    int i;
    int rc;
    unsigned char  CmdBuff[1024];

    //*****************************************************************
    // SEND THE 4.9.1.1.10 READ BURNED IN MAC ADDR COMMAND            *
    //*****************************************************************
    memset(CmdBuff, 0, 1024 );

    CmdBuff[0]  = (unsigned char)((12 & 0xFF00)>>8 ); // USB pkt length (LSB)
    CmdBuff[1]  = (unsigned char)( 12 & 0xFF);        // USB pkt length (MSB)
    CmdBuff[2]  = SP_READ;                               // Read
    CmdBuff[3]  = 0x05;                               // Command length
    CmdBuff[4]  = 0x00;                               // Data length LSB
    CmdBuff[5]  = 0x00;                               // Data Length MSB
    CmdBuff[6]  = 0x00;                               // Status
    CmdBuff[7]  = 0x00;                               // Reserved
    CmdBuff[8]  = 0x04;                               // 1st cmd byte
    CmdBuff[9]  = 0x09;                               // 2nd cmd byte
    CmdBuff[10] = 0x01;                               // 3rd cmd byte
    CmdBuff[11] = 0x01;                               // 4th cmd byte
    CmdBuff[12] = 0x0a;                               // 5th cmd byte
    CmdBuff[13] = usb_cksum(CmdBuff, 11);             // USB check sum 11;


    pSpSharedMem->CmdBuff[0] = 4;
    pSpSharedMem->CmdBuff[1] = 9;
    pSpSharedMem->CmdBuff[2] = 1;
    pSpSharedMem->CmdBuff[3] = 1;
    pSpSharedMem->CmdBuff[4] = 0x0a;

    if(debug) syslog(LOG_CRIT, "GetMaAddr(), Sending 4.9.1.1.10 command to the SP.\n");

    rc = SendCmdToSP(pSpSharedMem, pDevNode, CmdBuff, 14, MAX_RETRYS);

    if (rc == RC_SUCCESS)
    {
	    //if(debug) syslog(LOG_CRIT, "GetMaAddr(), 4.9.1.1.10 command returned success.\n");

	    memset(pDevNode->MacAddr, 0, MAC_ADDR_LEN);

	    for (i = 0; i < (int)CmdBuff[4]; i++)
	    {
	        pDevNode->MacAddr[i] = CmdBuff[13+i];
	    }

        //if(debug) syslog(LOG_CRIT, "GetMacAddr() Burned In Mac Address = %x %x %x %x %x %x\n",
        //                  pSpSharedMem->SPDevNode[node].MacAddr[0]
        //                  pSpSharedMem->SPDevNode[node].MacAddr[1]
        //                  pSpSharedMem->SPDevNode[node].MacAddr[2]
        //                  pSpSharedMem->SPDevNode[node].MacAddr[3]
        //                  pSpSharedMem->SPDevNode[node].MacAddr[4]
        //                  pSpSharedMem->SPDevNode[node].MacAddr[5] );
        if(debug) syslog(LOG_CRIT, "GetMacAddr() Burned In Mac Address = %x %x %x %x %x %x\n",
                         CmdBuff[13],
                         CmdBuff[14],
                         CmdBuff[15],
                         CmdBuff[16],
                         CmdBuff[17],
                         CmdBuff[18] );
    }

    return rc;
}





int GetNodeMode(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode)
{
    int i;
    int rc;
    unsigned char  CmdBuff[1024];

    pDevNode->NodeMode = PT_INVALID_MODE;

    //*****************************************************************
    // SEND THE 20.1.3.3.4 (Hermes) or 20.2.1.5 (Athena) READ NODE MODE COMMAND
    //*****************************************************************
    // The 20.1.3.3.4 command has recently (11/2007) become a Hermes-only dot command,
    // so it is not supported on Athena.  If issued to an Athena, the RSA II firmware 
    // will set the Status byte to 07, to indicate the command is UNSUPPORTED.
    memset(CmdBuff, 0, 1024 );

    CmdBuff[0]  = (unsigned char)((12 & 0xFF00)>>8 ); // USB pkt length (LSB)
    CmdBuff[1]  = (unsigned char)( 12 & 0xFF);        // USB pkt length (MSB)
    CmdBuff[2]  = SP_READ;                            // Read
    CmdBuff[3]  = 0x05;                               // Command length
    CmdBuff[4]  = 0x00;                               // Data length LSB
    CmdBuff[5]  = 0x00;                               // Data Length MSB
    CmdBuff[6]  = 0x00;                               // Status
    CmdBuff[7]  = 0x00;                               // Reserved
    CmdBuff[8]  = 0x14;                               // 1st cmd byte
    CmdBuff[9]  = 0x01;                               // 2nd cmd byte
    CmdBuff[10] = 0x03;                               // 3rd cmd byte
    CmdBuff[11] = 0x03;                               // 4th cmd byte
    CmdBuff[12] = 0x04;                               // 5th cmd byte
    CmdBuff[13] = usb_cksum(CmdBuff, 11);             // USB check sum 11;


    pSpSharedMem->CmdBuff[0] = 20;
    pSpSharedMem->CmdBuff[1] = 1;
    pSpSharedMem->CmdBuff[2] = 3;
    pSpSharedMem->CmdBuff[3] = 3;
    pSpSharedMem->CmdBuff[4] = 4;   

    if(debug) syslog(LOG_CRIT, "GetNodeMode(), Sending 20.1.3.3.4 command to the SP.\n");

    rc = SendCmdToSP(pSpSharedMem, pDevNode, CmdBuff, 14, MAX_RETRYS);

    if ( (rc == RC_SUCCESS) && 
         ( CmdBuff[2] == SP_CMDRESP ) && 
         ( CmdBuff[4] == 1 ) && 
         ( CmdBuff[6] == 0 ) )
    {
	    pDevNode->NodeMode = CmdBuff[13];

        if(debug) syslog(LOG_CRIT, "GetNodeMode() 20.1.3.3.4 command successful, Mode = %x\n", pDevNode->NodeMode );
    }
    else
    {
        if(debug) syslog(LOG_CRIT, "GetNodeMode(), 20.1.3.3.4 command failed. rc=%d, CmdType=%d, Status=%d, DataLen=%d \n",
                         rc, CmdBuff[2], CmdBuff[6], CmdBuff[4]);
        
        // Try issuing the Athena version of the same command, which is 20.2.1.5.
        CmdBuff[0]  = (unsigned char)((12 & 0xFF00)>>8 ); // USB pkt length (LSB)
        CmdBuff[1]  = (unsigned char)( 12 & 0xFF);        // USB pkt length (MSB)
        CmdBuff[2]  = SP_READ;                            // Read
        CmdBuff[3]  = 0x04;                               // Command length
        CmdBuff[4]  = 0x00;                               // Data length LSB
        CmdBuff[5]  = 0x00;                               // Data Length MSB
        CmdBuff[6]  = 0x00;                               // Status
        CmdBuff[7]  = 0x00;                               // Reserved
        CmdBuff[8]  = 0x14;                               // 1st cmd byte
        CmdBuff[9]  = 0x02;                               // 2nd cmd byte
        CmdBuff[10] = 0x01;                               // 3rd cmd byte
        CmdBuff[11] = 0x05;                               // 4th cmd byte
        CmdBuff[12] = usb_cksum(CmdBuff, 10);             // USB check sum 10;

        pSpSharedMem->CmdBuff[0] = 20;
        pSpSharedMem->CmdBuff[1] = 2;
        pSpSharedMem->CmdBuff[2] = 1;
        pSpSharedMem->CmdBuff[3] = 5;
        pSpSharedMem->CmdBuff[4] = 0;   

        if(debug) syslog(LOG_CRIT, "GetNodeMode(), Sending 20.2.1.5 command to the SP.\n");

        rc = SendCmdToSP(pSpSharedMem, pDevNode, CmdBuff, 14, MAX_RETRYS);

        if ( (rc == RC_SUCCESS) && 
             ( CmdBuff[2] == SP_CMDRESP ) && 
             ( CmdBuff[4] == 1 ) && 
             ( CmdBuff[6] == 0 ) )
        {
            pDevNode->NodeMode = CmdBuff[12];

            if(debug) syslog(LOG_CRIT, "GetNodeMode() 20.2.1.5 command successful Mode = %x\n", pDevNode->NodeMode );
        }
        else
        {
            if(debug) syslog(LOG_CRIT, "GetNodeMode(), 20.2.1.5 command failed. rc=%d, CmdType=%d, Status=%d, DataLen=%d \n",
                             rc, CmdBuff[2], CmdBuff[6], CmdBuff[4]);
        }
    }

    return rc;
}




int SendOSisDown(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode)
{
     int rc;
     unsigned char  CmdBuff[1024];

     //*****************************************************************
     // SEND THE 4.9.1.1.10 READ BURNED IN MAC ADDR COMMAND            *
     //*****************************************************************
     memset(CmdBuff, 0, 1024 );

     CmdBuff[0]  = (unsigned char)((11 & 0xFF00)>>8 ); // USB pkt length (LSB)
     CmdBuff[1]  = (unsigned char)( 11 & 0xFF);        // USB pkt length (MSB)
     CmdBuff[2]  = SP_WRITE;                           // Write command
     CmdBuff[3]  = 0x03;                               // Command Length
     CmdBuff[4]  = 0x01;                               // Data Length  LSB
     CmdBuff[5]  = 0x00;                               // Data Length  MSB
     CmdBuff[6]  = 0x00;                               // Status
     CmdBuff[7]  = 0x00;                               // Reserved
     CmdBuff[8]  = 0x04;                               // Command Byte 0
     CmdBuff[9]  = 0x03;                               // Command Byte 1
     CmdBuff[10] = 0x06;                               // Command Byte 2
     CmdBuff[11] = 0x04;                               // Data Byte 0
     CmdBuff[12] = usb_cksum(CmdBuff, 10);             // USB check sum 11;


     pSpSharedMem->CmdBuff[0] = 4;
     pSpSharedMem->CmdBuff[1] = 3;
     pSpSharedMem->CmdBuff[2] = 6;

     //if(debug) syslog(LOG_CRIT, "SendOSisDown() Sending 4.3.6 command to the SP.\n");

     rc = SendCmdToSP(pSpSharedMem, pDevNode, CmdBuff, 13, MAX_RETRYS);

     //if (rc == RC_SUCCESS)
     //{
	  //if(debug) syslog(LOG_CRIT, "SendOSisDown() 4.3.6 command returned success.\n");
     //}
     //else
     //{
     //if(debug) syslog(LOG_CRIT, "SendOSisDown() 4.3.6 command returned failure.\n");
     //}

     return rc;
}



int SendCmdToSP(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode, unsigned char *pBuffer, int BuffLen, int Retrys)
{
    int rc;
    int TimeOut;
    int BytesRead;
    int BytesWritten;
    int FirstPkt;
    int BytesToWrite;
    int BytesToRead;
    int BytesToCopy;
	 int RetryCount;

    unsigned char  TempBuff[4096];
    unsigned char *TempPtr;

    rc           = RC_TIMEOUT;
    RetryCount   = 0;				
    TempPtr      = pBuffer;
    BytesToWrite = BuffLen;

    //if(debug) syslog(LOG_CRIT, "SendCmdToSP(), Sending command to the SP, Cmd length = %i.\n", BuffLen);

    // SEND THE COMMAND TO THE SP ADAPTER VIA USB BULK WRITE
    while(BytesToWrite > 0)
    {
        BytesWritten = pSpSharedMem->sp_usb_bulk_write(pDevNode->pUsbHandle, 
                                                       pDevNode->OutEndPoint,
                                                       TempPtr, 
                                                       MAX_WRITE_LEN, 
                                                       3000);    // PHR_LINK

        // IF A WRITE FAILURE OCCURS START OVER
        if (BytesWritten != MAX_WRITE_LEN)
        {
            BytesToWrite = BuffLen;
            TempPtr  = pBuffer;
            RetryCount++;

            if (RetryCount >= Retrys)
            {
                //if(debug) syslog(LOG_CRIT, "SendCmdToSP(), Setting Reset flag due to command timeout.\n");
                return RC_WRITE_ERROR;
            }
        }
        else
        {
            BytesToWrite -= MAX_WRITE_LEN;
            TempPtr += MAX_WRITE_LEN;
            rc = RC_SUCCESS;
        }
    }

    //if(debug) syslog(LOG_CRIT, "SendCmdToSP(), Send command succeded, waiting for CmdResp.\n");

    // GET THE COMMAND RESPONSE FROM THE SP VIA USB BULK READ
    FirstPkt    = TRUE;
    BytesToRead = 1;          // initializer
    TempPtr     = TempBuff;
	 RetryCount  = 0;

    while(BytesToRead > 0)
    {
        BytesRead = 0;

        while(BytesRead != MAX_READ_LEN)
	     {
	        BytesRead = pSpSharedMem->sp_usb_bulk_read(pDevNode->pUsbHandle, 
                                                      pDevNode->InEndPoint,                   
                                                      TempPtr, 
                                                      MAX_READ_LEN, 
                                                      5000);   // PHR_LINK
           RetryCount++;

	        if (BytesRead != MAX_READ_LEN)
	        {
	            if (RetryCount >= Retrys)
	            {
	                return RC_READ_ERROR;
	            }
	        }
        }

        //if(debug) syslog(LOG_CRIT, "SendCmdToSP(), Read of CmdResp succeded, LoopCnt = %i.\n", LoopCnt);

        // IF THIS IS THE FIRST PACKET READ, PARSE THE HEADER
        if (FirstPkt)
        {
            FirstPkt = FALSE;

            BytesToRead = ScanRespPkt(pSpSharedMem, TempPtr);

            if(BytesToRead == -1)
            {
	            FirstPkt  = TRUE;
                BytesRead = MAX_READ_LEN;

		        // CLEAR OUT ANY OUTSTANDING DATA THAT MAY BE IN THE READ BUFFER
		        while(BytesRead == MAX_READ_LEN)
	            {
		            BytesRead = pSpSharedMem->sp_usb_bulk_read(pDevNode->pUsbHandle, 
                                                             pDevNode->InEndPoint,                    
                                                             TempPtr, 
                                                             MAX_READ_LEN, 
                                                             5000);   // PHR_LINK
                    RetryCount++;

		            if (RetryCount >= Retrys)
		            {
                        return RC_READ_ERROR;
		            }
	            }

                return RC_READ_ERROR;
            }
            else
            {
                BytesToCopy = BytesToRead;
            }
        }

        BytesToRead -= BytesRead;
        TempPtr += BytesRead;
    }

    memcpy(pBuffer, TempBuff, BytesToCopy);

    return RC_SUCCESS;
}


int  CountSPs(PSP_SHARED_MEM pSpSharedMem)            // PHR_LINK
{
   int SpCount;

   struct usb_bus      *pUsbBus;
   struct usb_bus      *pUsbBusses;                   // PHR_LINK
   struct usb_device   *pUsbDev;

   pSpSharedMem->sp_usb_init();                       // PHR_LINK
        //if(debug) syslog(LOG_CRIT, "called init\n");

   pSpSharedMem->sp_usb_find_busses();                // PHR_LINK
        //if(debug) syslog(LOG_CRIT, "called find busses\n");

   pSpSharedMem->sp_usb_find_devices();               // PHR_LINK
        //if(debug) syslog(LOG_CRIT, "called find devices\n");

   pUsbBusses = pSpSharedMem->sp_usb_get_busses();    // PHR_LINK
        //if(debug) syslog(LOG_CRIT, "called get busses\n");

   SpCount = 0;

   // SCAN THE USB DEVICES UNTIL WE FIND THIS PARICULAR USB DEVICE
   for ( pUsbBus = pUsbBusses; pUsbBus; pUsbBus = pUsbBus->next)         // PHR_LINK
   {
      //if(debug) syslog(LOG_CRIT, "found a bus\n");
      for (pUsbDev = pUsbBus->devices; pUsbDev; pUsbDev = pUsbDev->next)
      {
         //if(debug) syslog(LOG_CRIT, "found a dev\n");
         if(  (pUsbDev->descriptor.idVendor  == USB_VENDOR_ID  )  &&
             ((pUsbDev->descriptor.idProduct == USB_PRODUCT_ID )  ||
              (pUsbDev->descriptor.idProduct == USB_PRODUCT_ID2))  )
         {
             //if(debug) syslog(LOG_CRIT, "found a raptor\n");
             SpCount++;	
         }
      }
   }

   return SpCount;
}



// *********************************************************************
//  InitNode - THIS ROUTINE SCANS FOR A SPECIFIC RSA II DEVICE (NODE), *
//             INITIALIZES THE SPs NODE INFO, GETS A HANDLE FROM THE   *
//             USB LIB AND LOCKS THE USB DEVICE.                       *
// *********************************************************************
int InitNode(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pTempDevNode, int node)   // PHR_LINK
{
   struct usb_bus    *pUsbBus;
   struct usb_device *pUsbDev;
   struct usb_bus    *pUsbBusses;                              // PHR_LINK

   int i;            // DDH001
   int in_found=0;   // DDH001
   int out_found=0;  // DDH001
   int found=0;      // DDH001
   int wincfg=0;      // DDH002
   int detach_failed=0; // DDH002
   int UsbRC;
   int CurrNode;

   pSpSharedMem->sp_usb_init();                                // PHR_LINK

   pSpSharedMem->sp_usb_find_busses();                         // PHR_LINK

   pSpSharedMem->sp_usb_find_devices();                        // PHR_LINK

   pUsbBusses = pSpSharedMem->sp_usb_get_busses();             // PHR_LINK

   CurrNode = 0;

   // SCAN THE USB DEVICES UNTIL WE FIND THIS PARICULAR USB DEVICE
   for ( pUsbBus = pUsbBusses; pUsbBus; pUsbBus = pUsbBus->next)         // PHR_LINK
   {
      for ( pUsbDev = pUsbBus->devices; pUsbDev; pUsbDev = pUsbDev->next)
      {
	      if(  (pUsbDev->descriptor.idVendor  == USB_VENDOR_ID  )  &&
	          ((pUsbDev->descriptor.idProduct == USB_PRODUCT_ID )  ||
	           (pUsbDev->descriptor.idProduct == USB_PRODUCT_ID2))  )
         {
	         if (node == CurrNode)
            {
 	            pTempDevNode->pUsbDev = pUsbDev;
                    // ISSUE THE USB_OPEN
	            pTempDevNode->pUsbHandle = pSpSharedMem->sp_usb_open(pTempDevNode->pUsbDev);   // PHR_LINK
	            if (pTempDevNode->pUsbHandle == NULL )
	            {
	               if (debug) syslog(LOG_CRIT, "RSA II node %i failed to open, returning failure.\n", node);
	               pTempDevNode->pUsbHandle = (void *)INVALID_USB_HANDLE;
	               return RC_SYSTEM_ERROR;
	            }
	            else
	            {
	               if (debug) syslog(LOG_CRIT, "RSA II node %i successfully opened.\n", node);
	               pTempDevNode->State = NODE_OPENED;
	            }

               //  DDH001  Scan thru configs/interfaces to find the right interface to claim
               //  DDH001  Then save the interface num, in and out endpoints
               for (i = 0; i < pUsbDev->descriptor.bNumConfigurations; i++) 
               {
                  struct usb_config_descriptor *config_desc;
                  int j; 
                  config_desc = &pUsbDev->config[i];

                  if(debug) syslog(LOG_CRIT, "NumInterfaces = %i.\n", config_desc->bNumInterfaces );  // DBG
                  for (j = 0; j < config_desc->bNumInterfaces; j++) 
                  {
                     struct usb_interface *interface;
                     int k; 

                     interface = &config_desc->interface[j];

                     if(debug) syslog(LOG_CRIT, "num_altsetting = %i.\n", interface->num_altsetting );  // DBG
                     for (k = 0; k < interface->num_altsetting; k++) 
                     {
                        struct usb_interface_descriptor *interface_desc;
                        int l; 
                        
                        interface_desc = &interface->altsetting[k]; 

                        if ( ( interface_desc->bInterfaceClass == 255 ) &&
                             ( interface_desc->bInterfaceSubClass == 0 ) &&
                             ( interface_desc->bInterfaceProtocol == 0 ) ) 
                        {
                           if(debug) syslog(LOG_CRIT, "NumEndpoints = %i.\n", interface_desc->bNumEndpoints );  // DBG

                           pTempDevNode->UsbInterface= interface_desc->bInterfaceNumber;

                           for (l = 0; l < interface_desc->bNumEndpoints; l++) 
                           {
                              struct usb_endpoint_descriptor *endpoint;
                              endpoint = &interface_desc->endpoint[l];
                              //endpoint->wMaxPacketSize); 
               
                              if(debug) syslog(LOG_CRIT, "EndpointAddress = %x.\n", endpoint->bEndpointAddress );  // DBG
                              if ( ( (endpoint->bEndpointAddress) &0x80) == 0x80 )
                              {
                                  pTempDevNode->InEndPoint= endpoint->bEndpointAddress;
                                  in_found = 1;
                              }
                              else
                              {
                                  pTempDevNode->OutEndPoint= endpoint->bEndpointAddress;
                                  out_found = 1;
                              }
                           }
                           if (( in_found != 0 ) && (out_found != 0 )) 
                           {
                              found=1;
                           }
                        }
                        if ( ( interface_desc->bInterfaceClass == 3 ) &&
                             ( interface_desc->bInterfaceSubClass == 0 ) &&
                             ( interface_desc->bInterfaceProtocol == 0 ) ) 
                        {
                           wincfg=1;  // found a windows-configured RSA2

                           if ( pSpSharedMem->sp_usb_detach_kernel_driver_np != -1 ) 
                           {
                              if(debug) syslog(LOG_CRIT, "Attempting to remove kernel HID driver from SP interface.\n" );  // DBG

                              UsbRC = pSpSharedMem->sp_usb_detach_kernel_driver_np( pTempDevNode->pUsbHandle, interface_desc->bInterfaceNumber );
                              
                              if (UsbRC == 0) 
                              {
                                 detach_failed=1;
                              }
                           }

                           if(debug) syslog(LOG_CRIT, "NumEndpoints = %i.\n", interface_desc->bNumEndpoints );  // DBG
      
                           pTempDevNode->UsbInterface= interface_desc->bInterfaceNumber;
      
                           for (l = 0; l < interface_desc->bNumEndpoints; l++) 
                           {
                              struct usb_endpoint_descriptor *endpoint;
                              endpoint = &interface_desc->endpoint[l];
                              //endpoint->wMaxPacketSize); 
                     
                              if(debug) syslog(LOG_CRIT, "EndpointAddress = %x.\n", endpoint->bEndpointAddress );  // DBG
                              if ( ( (endpoint->bEndpointAddress) &0x80) == 0x80 )
                              {
                                  pTempDevNode->InEndPoint= endpoint->bEndpointAddress;
                                  in_found = 1;
                              }
                              else
                              {
                                  pTempDevNode->OutEndPoint= endpoint->bEndpointAddress;
                                  out_found = 1;
                              }
                           }
                           if (( in_found != 0 ) && (out_found != 0 )) 
                           {
                              found=1;
                           }
                        }
                        if (found==1) break;
                     }
                     if (found==1) break;
                  }
                  if (found==1) break;
               }

               if (found==1) 
               {
	               if(debug) syslog(LOG_CRIT, "InitNode: claiming interface handle = %x, interface = %x.\n", pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );
                  // CLAIM THE USB INTERFACE
                  UsbRC = pSpSharedMem->sp_usb_claim_interface(pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );       // PHR_LINK
                  if (UsbRC != RC_SUCCESS)
                  {

                     if (wincfg==1) // found a windows-configured RSA2
                     {
                        // if we can't claim a win config, give up
                        if ( pSpSharedMem->InitialLoad == TRUE ) syslog(LOG_CRIT, "The RSA II is not properly configured for Linux.\n" );  
                        return RC_SYSTEM_ERROR;
                     }
                     else 
                     {
                        if ( pSpSharedMem->sp_usb_detach_kernel_driver_np != -1 ) {
                            UsbRC = pSpSharedMem->sp_usb_detach_kernel_driver_np( pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );
                            if (UsbRC != RC_SUCCESS) {
                                if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_detach_kernel_driver_np() failed for node %i.\n", node);  // PHR_LINK
                            }
                            else if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_detach_kernel_driver_np() succeeded for node %i.\n",node); // PHR_LINK
                            // TRY TO CLAIM INTERFACE AGAIN
                            UsbRC = pSpSharedMem->sp_usb_claim_interface(pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );       // PHR_LINK
                            if (UsbRC == RC_SUCCESS) {
                                    if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_claim_interface() succeeded for node %i.\n",node); // PHR_LINK
                                    return RC_SUCCESS;
                            }
                            else if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_claim_interface() failed for node %i.\n", node);    // PHR_LINK
                        }

                        // SET ACTIVE CONFIGURATION
                        UsbRC = pSpSharedMem->sp_usb_set_configuration(pTempDevNode->pUsbHandle, USB_SP_CONFIG);         // PHR_LINK
                        if (UsbRC != RC_SUCCESS)
                        {
                           if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_set_configuration() failed for node %i.\n", node);  // PHR_LINK
                           pSpSharedMem->sp_usb_close(pTempDevNode->pUsbHandle);                                        // PHR_LINK
                           pTempDevNode->State      = NODE_AVAILABLE;
                           pTempDevNode->pUsbHandle = (void *)INVALID_USB_HANDLE;
                           return RC_SYSTEM_ERROR;
                        }
                        if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_set_configuration() succeeded for node %i.\n",node);    // PHR_LINK
                        // TRY CLAIM THE USB INTERFACE AGAIN
                        UsbRC = pSpSharedMem->sp_usb_claim_interface(pTempDevNode->pUsbHandle, pTempDevNode->UsbInterface );       // PHR_LINK
                        if (UsbRC != RC_SUCCESS)
                        {
                           if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_claim_interface() failed for node %i.\n", node);    // PHR_LINK
                           pSpSharedMem->sp_usb_close(pTempDevNode->pUsbHandle);                                        // PHR_LINK
                           pTempDevNode->State      = NODE_AVAILABLE;
                           pTempDevNode->pUsbHandle = (void *)INVALID_USB_HANDLE;
                           return RC_SYSTEM_ERROR;
                        }
                        if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_claim_interface() succeeded for node %i.\n",node); // PHR_LINK
                        return RC_SUCCESS;
                     }
                  }
                  if(debug) syslog(LOG_CRIT, "pSpSharedMem->sp_usb_claim_interface() succeeded for node %i.\n",node); // PHR_LINK
                  return RC_SUCCESS;
               }
               else return RC_SYSTEM_ERROR;
            }
	         else
	         {
               CurrNode++;
            }
         }
      }
   }

   return RC_SYSTEM_ERROR;
}

int MACcmp(const char *s1, const char *s2)
{
    int i;
    int rc=0;
    
    for( i=0; i<6; i++ )
    {
       if (s1[i] != s2[i])
       {
          rc = -1;
          break;
       }
    }
    return rc;
}


int AddNode(PSP_SHARED_MEM pSpSharedMem, PSP_DEV_NODE pDevNode)
{
    int i;
    int rc;
    int node;
    int Found = FALSE;

     if(debug) syslog(LOG_CRIT, "AddNode(): Mode = %x\n", pDevNode->NodeMode);   
    // ADD THE NODE IF THIS IS AN INITIAL LOAD OF IBMASM
    if (pSpSharedMem->InitialLoad)
    {
        if ( ( pDevNode->NodeMode == PT_INVALID_MODE ) ||
             ( pDevNode->NodeMode == PT_NULL_MODE ) ||
             ( pDevNode->NodeMode == PT_PRIMARY_MODE ) ||
             ( pDevNode->NodeMode == PT_STANDALONE_MODE ) ||
             ( pDevNode->NodeMode == PT_STANDALONE_PRIMARY_MODE ) ||
             ( pDevNode->NodeMode == PT_STANDALONE_RESET_PRIMARY_MODE ) )
        {
            // This is the primary node, so make it node 0
            if(debug) syslog(LOG_CRIT, "AddNode() is adding a new primary node, Mode = %x", pDevNode->NodeMode);
            node = 0;
            Found = TRUE;
            rc = RC_SUCCESS;
        }
        else
        {

            for (node = 1; node < MAX_SP_NODES; node++ )
            {
               if (pSpSharedMem->SPDevNode[node].State == NODE_AVAILABLE)
               {
                   break;
               }
            }
         
            if (pSpSharedMem->SPDevNode[node].State == NODE_AVAILABLE)
            {
               if(debug) syslog(LOG_CRIT, "AddNode() is adding a new secondary node, Node = %i", node);
               Found = TRUE;
               rc = RC_SUCCESS;
            }
            else
            {
               if(debug) syslog(LOG_CRIT, "AddNode(): 1st load, node # not found \n");   
               Found = FALSE;
               rc = RC_SYSTEM_ERROR;
            }
        }
    }
    else
    {
        // LOOK FOR A MATCHING NODE THAT WAS PREVIOUSY CONNECTED
        for (node = 0; node < MAX_SP_NODES; node++ )
        {
            if (pSpSharedMem->SPDevNode[node].State != NODE_AVAILABLE)   // PHR_225158
            {
                rc = MACcmp(pSpSharedMem->SPDevNode[node].MacAddr, pDevNode->MacAddr);

                if (rc == 0)     // strcmp returns 0 if strings match
                {
                    if(debug) syslog(LOG_CRIT, "AddNode(): Match MAC, node = %x, SM-MAC[5]=%x, tmpMAC[5]=%x, state=%x\n", node, pSpSharedMem->SPDevNode[node].MacAddr[5], pDevNode->MacAddr[5], pSpSharedMem->SPDevNode[node].State );   

                    if(pSpSharedMem->SPDevNode[node].State == NODE_RESETTING)
                    {
                        pSpSharedMem->ResetEventInfo.node  = node;
                        pSpSharedMem->ResetEventInfo.event = EVENT_RESET_COMPLETE;
                        syslog(LOG_CRIT, "RSA II device support restored for node %i, SP reset completed successfully.\n", node);

                    }
                    else if(pSpSharedMem->SPDevNode[node].State == NODE_DISCONNECTED)
                    {
                        pSpSharedMem->ResetEventInfo.node  = node;
                        pSpSharedMem->ResetEventInfo.event = EVENT_RECONNECT;
                        syslog(LOG_CRIT, "RSA II device support restored for node %i, node has been reconnected.\n", node);
                    }

                    Found = TRUE;
            	     rc = RC_SUCCESS;
                    break;
                }
            }
        }


        // IF THIS IS NOT A RE-CONNECTION, ASSIGN IT TO THE FIRST NODE AVAILABLE
        if (!Found)
        {
           if ( ( pDevNode->NodeMode == PT_INVALID_MODE ) ||
                ( pDevNode->NodeMode == PT_NULL_MODE ) ||
                ( pDevNode->NodeMode == PT_PRIMARY_MODE ) ||
                ( pDevNode->NodeMode == PT_STANDALONE_MODE ) ||
                ( pDevNode->NodeMode == PT_STANDALONE_PRIMARY_MODE ) ||
                ( pDevNode->NodeMode == PT_STANDALONE_RESET_PRIMARY_MODE ) )
           {
               // This is the primary node, so make it node 0
               if(debug) syslog(LOG_CRIT, "AddNode() is adding a new primary node, Mode = %x", pDevNode->NodeMode );
               node = 0;
               Found = TRUE;
               rc = RC_SUCCESS;
           }
           else
           {
              for (node = 1; node < MAX_SP_NODES; node++ )
              {
                  if (pSpSharedMem->SPDevNode[node].State == NODE_AVAILABLE)
                 {
                     if(debug) syslog(LOG_CRIT, "AddNode() adding new node node = %i,  Mode = %x\n", node, pDevNode->NodeMode);   
                     Found = TRUE;
                     rc = RC_SUCCESS;
                     break;
                  }
              }
           }
        }
    }


    if(Found)
    {
        //if(debug) syslog(LOG_CRIT, "Adding node = %x, state = %x.\n", node, pDevNode->State );
	     pSpSharedMem->SPDevNode[node].State       = pDevNode->State;
    	  pSpSharedMem->SPDevNode[node].pUsbDev     = pDevNode->pUsbDev;
        pSpSharedMem->SPDevNode[node].OutEndPoint = pDevNode->OutEndPoint;
    	  pSpSharedMem->SPDevNode[node].InEndPoint  = pDevNode->InEndPoint;
    	  pSpSharedMem->SPDevNode[node].pUsbHandle  = pDevNode->pUsbHandle;
    	  pSpSharedMem->SPDevNode[node].UsbInterface= pDevNode->UsbInterface;

        memset(pSpSharedMem->SPDevNode[node].MacAddr, 0, MAC_ADDR_LEN);

    	  for (i = 0; i < MAC_ADDR_LEN; i++)
    	  {
    	     pSpSharedMem->SPDevNode[node].MacAddr[i] = pDevNode->MacAddr[i];
    	  }

        if(debug) syslog(LOG_CRIT, "AddNode() mac address of node %i =  %x %x %x %x %x %x\n.",
                          node,
                          pSpSharedMem->SPDevNode[node].MacAddr[0],
                          pSpSharedMem->SPDevNode[node].MacAddr[1],
                          pSpSharedMem->SPDevNode[node].MacAddr[2],
                          pSpSharedMem->SPDevNode[node].MacAddr[3],
                          pSpSharedMem->SPDevNode[node].MacAddr[4],
                          pSpSharedMem->SPDevNode[node].MacAddr[5] );

    }
    else
    {
	    //if(debug) syslog(LOG_CRIT, "Could not add RSA II node, MAX_DEV_NODES exceeded.\n");
	    rc = RC_SYSTEM_ERROR;
    }

    return rc;
}


int EnableShutdownEvent(PSP_SHARED_MEM pSpSharedMem, int node)
{
    int rc;
    unsigned char Buffer[1024];

    PSP_DEV_NODE pDevNode;

    pDevNode = &pSpSharedMem->SPDevNode[node];

    //if(debug) syslog(LOG_CRIT, "EnableShutdownEvent(), enabling shutdown event for node %i. \n", node);

    memset(Buffer, 0, 1024);

    Buffer[0] = (unsigned char)((11 & 0xFF00)>>8 ); // USB pkt length (LSB)
    Buffer[1] = (unsigned char)( 11 & 0xFF);        // USB pkt length (MSB)

    Buffer[2]  = SP_WRITE;                          // Command Type
    Buffer[3]  = 3;                                 // Command Length
    Buffer[4]  = 1;                                 // Data Length  LSB
    Buffer[5]  = 0;                                 // Data Length  MSB
    Buffer[6]  = 0;                                 // Status
    Buffer[7]  = 0;                                 // Reserved
    Buffer[8]  = 5;                                 // Command Byte 0
    Buffer[9]  = 2;                                 // Command Byte 1
    Buffer[10] = 4;                                 // Command Byte 2
    Buffer[11] = 1;                                 // Data Byte 0

    Buffer[12] = usb_cksum(Buffer, 10);           // USB check sum

    pSpSharedMem->CmdBuff[0] = 5;
    pSpSharedMem->CmdBuff[1] = 2;
    pSpSharedMem->CmdBuff[2] = 4;

    //if(debug) syslog(LOG_CRIT, "EnableShutdownEvent(), sending event enable command 5.2.4 to the SP.\n");

    rc = SendCmdToSP(pSpSharedMem, pDevNode, Buffer, 13, MAX_RETRYS);

    if (rc != RC_SUCCESS)
    {
	    //if(debug) syslog(LOG_CRIT, "EnableShutdownEvent(), Command 5.2.4 failed, returning failure.\n");
	    return RC_SYSTEM_ERROR;
    }
    else
    {
	    //if(debug) syslog(LOG_CRIT, "EnableShutdownEvent(), Command 5.2.4 returned successfully.\n");
	    return RC_SUCCESS;
    }
}


// added by PHR_225158
void process_reset_event(PSP_SHARED_MEM pSpSharedMem, int node, int event)
{
    switch (event)
    {
       case EVENT_RESET_COMPLETE:
          //if(debug) syslog(LOG_CRIT,  "process_reset_event() setting EVENT_RESET_COMPLETE flags.\n");
          pSpSharedMem->ResetEventInfo.event = EVENT_RESET_COMPLETE;
          pSpSharedMem->ResetEventInfo.node  = node;
          break;
       case EVENT_DISCONNECT:
          //if(debug) syslog(LOG_CRIT,  "process_reset_event() setting EVENT_DISCONNECT flags.\n");
          pSpSharedMem->ResetEventInfo.event = EVENT_DISCONNECT;
          pSpSharedMem->ResetEventInfo.node  = node;
          break;
       case EVENT_RECONNECT:
          //if(debug) syslog(LOG_CRIT,  "process_reset_event() setting EVENT_RECONNECT flags.\n");
          pSpSharedMem->ResetEventInfo.event = EVENT_RECONNECT;
          pSpSharedMem->ResetEventInfo.node  = node;
          break;
    }

    return;
}



int InitDynamicLinks(PSP_SHARED_MEM pSpSharedMem)
{
    // if(debug) syslog(LOG_CRIT, "Initializing Dynamic Links.\n");
    const char *pError;

    // CLEAR ANY OUTSTANDING LINK ERRORS
    pError = dlerror();

    /*****************************************************************/
    /*  GET A HANDLE TO THE libusb.so LINK LIBRARY                   */
    /*****************************************************************/
    pSpSharedMem->pLib = dlopen ("libusb-0.1.so.4", RTLD_NOW);

    if (pSpSharedMem->pLib == NULL)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlopen of libusb.so failed dlerror = %s.\n", dlerror());

        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_init SYMBOL                         */
    /*****************************************************************/
    pSpSharedMem->sp_usb_init = dlsym(pSpSharedMem->pLib, "usb_init");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_init failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }

    /*****************************************************************/
    /*  GET A POINTER TO THE usb_close SYMBOL                        */
    /*****************************************************************/
    pSpSharedMem->sp_usb_close = dlsym(pSpSharedMem->pLib, "usb_close");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_close failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_find_busses SYMBOL                  */
    /*****************************************************************/
    pSpSharedMem->sp_usb_find_busses = dlsym(pSpSharedMem->pLib, "usb_find_busses");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_find_busses failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_find_devices SYMBOL                 */
    /*****************************************************************/
    pSpSharedMem->sp_usb_find_devices = dlsym(pSpSharedMem->pLib, "usb_find_devices");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_find_devices failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_bulk_write SYMBOL                   */
    /*****************************************************************/
    pSpSharedMem->sp_usb_bulk_write = dlsym(pSpSharedMem->pLib, "usb_bulk_write");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_bulk_write failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_bulk_read SYMBOL                    */
    /*****************************************************************/
    pSpSharedMem->sp_usb_bulk_read = dlsym(pSpSharedMem->pLib, "usb_bulk_read");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_bulk_read failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_control_msg SYMBOL                  */
    /*****************************************************************/
    pSpSharedMem->sp_usb_control_msg = dlsym(pSpSharedMem->pLib, "usb_control_msg");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_control_msg failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_set_configuration SYMBOL            */
    /*****************************************************************/
    pSpSharedMem->sp_usb_set_configuration = dlsym(pSpSharedMem->pLib, "usb_set_configuration");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_set_configuration failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_claim_interface SYMBOL              */
    /*****************************************************************/
    pSpSharedMem->sp_usb_claim_interface = dlsym(pSpSharedMem->pLib, "usb_claim_interface");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_claim_interface failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_release_interface SYMBOL            */
    /*****************************************************************/
    pSpSharedMem->sp_usb_release_interface = dlsym(pSpSharedMem->pLib, "usb_release_interface");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_release_interface failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_set_altinterface SYMBOL             */
    /*****************************************************************/
    pSpSharedMem->sp_usb_set_altinterface = dlsym(pSpSharedMem->pLib, "usb_set_altinterface");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_set_altinterface failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_resetep SYMBOL                      */
    /*****************************************************************/
    pSpSharedMem->sp_usb_resetep = dlsym(pSpSharedMem->pLib, "usb_resetep");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_resetep failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_clear_halt SYMBOL                   */
    /*****************************************************************/
    pSpSharedMem->sp_usb_clear_halt = dlsym(pSpSharedMem->pLib, "usb_clear_halt");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_clear_halt failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_reset SYMBOL                        */
    /*****************************************************************/
    pSpSharedMem->sp_usb_reset = dlsym(pSpSharedMem->pLib, "usb_reset");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_reset failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_get_busses SYMBOL                   */
    /*****************************************************************/
    pSpSharedMem->sp_usb_get_busses = dlsym(pSpSharedMem->pLib, "usb_get_busses");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_get_busses failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_open SYMBOL                         */
    /*****************************************************************/
    pSpSharedMem->sp_usb_open = dlsym(pSpSharedMem->pLib, "usb_open");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_open failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_device SYMBOL                       */
    /*****************************************************************/
    pSpSharedMem->sp_usb_device = dlsym(pSpSharedMem->pLib, "usb_device");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_device failed dlerror = %s.\n", dlerror());
        dlclose(pSpSharedMem->pLib);
        return RC_SYSTEM_ERROR;
    }


    /*****************************************************************/
    /*  GET A POINTER TO THE usb_device SYMBOL                       */
    /*****************************************************************/
    pSpSharedMem->sp_usb_detach_kernel_driver_np = dlsym(pSpSharedMem->pLib, "usb_detach_kernel_driver_np");

    pError = dlerror();

    if (pError)
    {
        //if(debug) syslog(LOG_CRIT, "LINKUSB: dlsym of usb_detach.... failed dlerror = %s.\n", dlerror());
        // not a fatal error, just flag link as missing
        pSpSharedMem->sp_usb_detach_kernel_driver_np = -1;
    }


    return RC_SUCCESS;
}


