// **********************************************************************
// * ibmasm.h                                                           *
// *                                                                    *
// * Written by Pete Reynolds                                           *
// * (C) Copyright IBM Corporation, 1998-2003                           *
// *                                                                    *
// *  This software may be used and distributed according to the terms  *
// *  of the GNU Public License, incorporated herein by reference.      *
// **********************************************************************

// **********************************************************************
// * ibmasm.h  - This header file includes all defines, includes and    *
// *             typdefs for the IBM Service Processor USB deamon.      *
// *                                                                    *
// * CHANGE HISTORY:                                                    *
// * Date      Change ID   Description                                  *
// * --------  ----------  -------------------------------------------  *
// * 11/18/02  P.Reynolds  Initial Creation                             *
// * --------  ----------  -------------------------------------------  *
// *                                                                    *
// **********************************************************************




// ***************************************
//   INCLUDES
// ***************************************
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <dlfcn.h>                    // PHR_LINK
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <usb.h>
#include <syslog.h>
#include <sys/stat.h>

#define USB_VENDOR_ID     0x04B3
#define USB_PRODUCT_ID    0x4001
#define USB_PRODUCT_ID2   0x4003


#define LINUX_DD_MAJOR    1
#define LINUX_DD_MINOR    08	 // FOR BUILD ibmusbasm-1.08-2.i386.rpm

#define TRUE  1
#define FALSE 0

// COMMAND STATES
#define CMD_IDLE          0
#define CMD_INITIATED     1
#define CMD_PENDING       2
#define CMD_COMPLETE      3
#define CMD_READ_ERR      4
#define CMD_WRITE_ERR     5

// I2C COMMAND EQUATES
#define SP_WRITE          0
#define SP_WRITENEXT      1
#define SP_READ           2
#define SP_READNEXT       3
#define SP_CMDRESP        4
#define SP_EVENT          5
#define SP_HEARTBEAT      6


#define MAX_RETRYS        5
#define MAX_CONFIGS       8
#define MAX_SP_NODES      16
#define MAX_HANDLES       256
#define EVENT_BUFF_SIZE   256
#define MAX_BUFF_SIZE	  4096
#define MAX_READ_LEN      64
#define MAX_WRITE_LEN     64
#define EVENT_QUEUE_DEPTH 16

#define CMD_BLK_LEN       6          // 6 bytes of command block
#define CHK_SUM_LEN       1          // 1 byte of usb check sum
#define USB_PKT_LEN       2          // 2 byte of usb packet length
#define MAC_ADDR_LEN      16	     // 10 bytes of Mac Address data + pad

#define USB_SP_CONFIG     1          // default configuration
#define USB_SP_INTERFACE  2          // default interface - vendor defined
#define USB_SP_ALT_INTERF 0          // default alternate interface
#define USB_SP_IN_EP      0x83
#define USB_SP_OUT_EP     0x3

// ******************************************************
// THIS IS AN ATTEMPT TO TRY TO GET UNIQUE KEYS        *
// ******************************************************
#define SHARED_MEM_KEY    (USB_VENDOR_ID+USB_PRODUCT_ID)
#define CMD_RESP_SEM_KEY  USB_VENDOR_ID
#define CMD_RCVD_SEM_KEY  USB_PRODUCT_ID



#define MEM_LOCK_FILE       "/var/lock/subsys/ibmusbasm"
#define ASM_PID_FILE        "/var/run/ibmusbasm.pid"

#define CMD_TIMEOUT         45
#define CMD_TIMEOUT_RESET  255     // Used while RSA II is resetting.  Particularly useful while all nodes
                                   // in a multi-node system reset and the daemon takes a long amount of
                                   // time to re-discover them all.


// *******************************************************
//  5 POSSIBLE STATES OF DEVICE NODES	  PHR_225158    *
// *******************************************************
#define NODE_AVAILABLE      0	// Device node is available
#define NODE_OPENED         1	// Device has been located and has a open handle
#define NODE_CLOSED	        2	// Device was closed on exit
#define NODE_RESETTING      3	// SP device is going through a reset
#define NODE_DISCONNECTED   4	// SP device was has been disconnected - USB cable unplugged


// *******************************************************
//  4 POSSIBLE RESET/CONNECTION EVENTS     PHR_225158    *
// *******************************************************
#define EVENT_RESET_START    0
#define EVENT_RESET_COMPLETE 1
#define EVENT_DISCONNECT     2
#define EVENT_RECONNECT      3


// *************************************************
//  ERROR / RETURN CODE DEFINITIONS                *
// *************************************************
#define RC_READ_FAILURE                          -1
#define INVALID_USB_HANDLE                       -1
#define IPC_FAILURE                              -1

#define RC_SUCCESS                                0
#define RC_INVALID_HANDLE                         1
#define RC_TIMEOUT                                2
#define RC_NOT_OPEN                               3
#define RC_OPEN_FAILED                            4
#define RC_CLOSE_FAILED                           5
#define RC_ALREADY_REGISTERED                     6
#define RC_EVENT_NOT_REGISTERED                   7
#define RC_SYSTEM_ERROR                           8
#define RC_INVALID_BUFFER                         9
#define RC_EVENT_CANCELLED                       10
#define RC_DRIVER_CLOSED                         11
#define RC_NO_EVENT_BFRS                         13
#define RC_ALREADY_OPEN                          14
#define RC_REVERSE_HB_ALREADY_REGISTERD          15
#define RC_REVERSE_HEARTBEAT_HAS_FAILED          17
#define RC_REVERSE_HB_ALREADY_STOPPED            18
#define RC_INTERRUPTED                           19
#define RC_RESET                                 20
#define RC_READ_ERROR                            21
#define RC_WRITE_ERROR                           22


// *************************************************
//  NODE MODE DEFINITIONS                          *
// *************************************************
#define PT_INVALID_MODE   0xFFFFFFFF
#define PT_NULL_MODE   0 
#define PT_PRIMARY_MODE   1 
#define PT_SECONDARY_MODE   2 
#define PT_STANDALONE_MODE   3 
#define PT_STANDALONE_RESET_PRIMARY_MODE   4 
#define PT_STANDALONE_RESET_SECONDARY_MODE   5 
#define PT_STANDALONE_PRIMARY_MODE   6 
#define PT_STANDALONE_SECONDARY_MODE   7

typedef int SYS_HANDLE;

int debug = FALSE;
// ************************************************************
//  BUFFER THAT CONTAINS 1 EVENT ENTRY - 256 BYTE MAX         *
// ************************************************************
typedef unsigned char EVENT[EVENT_BUFF_SIZE];


// Start of PHR_LINK
// ***************************************************************
// TYPDEFS REQUIRED FOR DYNAMICALLY LINKED USB FUNCTION CALLS   *
// ***************************************************************
typedef void (*SP_USB_INIT)              (void);
typedef int  (*SP_USB_CLOSE)             (usb_dev_handle *dev);
typedef int  (*SP_USB_FIND_BUSSES)       (void);
typedef int  (*SP_USB_FIND_DEVICES)      (void);
typedef int  (*SP_USB_BULK_WRITE)        (usb_dev_handle *dev, int ep, char *bytes, int size, int timeout);
typedef int  (*SP_USB_BULK_READ)         (usb_dev_handle *dev, int ep, char *bytes, int size, int timeout);
typedef int  (*SP_USB_CONTROL_MSG)       (usb_dev_handle *dev, int requesttype, int request, int value, int index, char *bytes, int size, int timeout);
typedef int  (*SP_USB_SET_CONFIGURATION) (usb_dev_handle *dev, int configuration);
typedef int  (*SP_USB_CLAIM_INTERFACE)   (usb_dev_handle *dev, int interface);
typedef int  (*SP_USB_RELEASE_INTERFACE) (usb_dev_handle *dev, int interface);
typedef int  (*SP_USB_SET_ALTINTERFACE)  (usb_dev_handle *dev, int alternate);
typedef int  (*SP_USB_RESETEP)           (usb_dev_handle *dev, unsigned int ep);
typedef int  (*SP_USB_CLEAR_HALT)        (usb_dev_handle *dev, unsigned int ep);
typedef int  (*SP_USB_RESET)             (usb_dev_handle *dev);
typedef int  (*SP_USB_DETACH)            (usb_dev_handle *dev, int interface); 

typedef struct usb_bus*        (*SP_USB_GET_BUSSES) (void);
typedef struct usb_dev_handle* (*SP_USB_OPEN)       (struct usb_device *dev);
typedef struct usb_device*     (*SP_USB_DEVICE)     (usb_dev_handle *dev);
//typedef char *sp_usb_strerror(void);
// Start of PHR_LINK


// ************************************************************
//  QUEUE THAT HOLDS 16 EVENTS                                *
// ************************************************************
typedef struct _SP_EVENT_QUEUE
{
   EVENT Event[16];

} SP_EVENT_QUEUE, *PSP_EVENT_QUEUE;



// ************************************************************
//  SERVICE PROCESSOR NODE INFO - ADAPTOR SPECIFIC            *
// ************************************************************
typedef struct _SP_DEV_NODE
{
   int            Handle;
   int            State;
   int            ErrorCount;
   int            BufferSize;
   int            OutEndPoint;
   int            InEndPoint;
   int            UsbInterface;  // DDH001
   int            EventIndex;
   int            NodeMode;
   SP_EVENT_QUEUE EventQueue;
   unsigned char  MacAddr[MAC_ADDR_LEN];

#ifdef __x86_64__
   struct usb_dev_handle  *pUsbHandle;
   struct usb_device      *pUsbDev;
#endif

//  NOTE!! These fields MUST NOT be accessed from 32-bit code
//  The fields are just placeholders to make the struct the right size
#ifdef __i386__
   struct usb_dev_handle  *pUsbHandle;
   struct usb_dev_handle  *pUsbHandle_PAD;
  
   struct usb_device      *pUsbDev;
   struct usb_device      *pUsbDev_PAD;
#endif

} SP_DEV_NODE, *PSP_DEV_NODE;

// ***********************************************************
// **********************************************************
typedef struct _EVENT_REG_INFO
{
   int Initialized;
   int Handle;
   int Waiting;
   int EventIndex;
   int Pid;
   int Node;
   int Deregister;
   int EventReceived;
} EVENT_REG_INFO, *PEVENT_REG_INFO;





// ************************************************************
//  LIST OF CALLING PROCESSES                                 *
// ***********************************************************
typedef struct _SP_CALLER_LIST
{
   int Pid;                           // Callers process ID
   int BufferSize;                    // Callers buffer size in bytes
   int ReturnCode;                    // return code of requested function
   int Node;                          // node that the caller issued open for
   int RegForRH;                      // flag indicating if callers is reg for RH

   EVENT_REG_INFO   EventRegInfo;     // Event Info

} SP_CALLER_LIST, *PSP_CALLER_LIST;


// ************************************************************
//  LIST OF LOCALY ISSUED HANDLES                             *
// ************************************************************
typedef struct _SP_HANDLE_LIST
{
   int Handle;
   int Free;

} SP_HANDLE_LIST, *PSP_HANDLE_LIST;


// ***********************************************
//  FORMAT OF A SERVICE PROCESSOR COMMAND PACKET *
// ***********************************************
typedef struct _SPCMD
{
   unsigned char      CommandType;         // Read, Write, Resp...
   unsigned char      CommandLgth;         // Length of the dot cmd
   unsigned short int DataLgth;            // Length fo the data
   unsigned char      Status;              // return code
   unsigned char      Reserved;            // reserved

} SPCMD, *PSPCMD;

// PHR_225158
// ************************************************
//  HOLDS DATA REQUIRED FOR ISSUEING RESET EVENTS *
// ************************************************
typedef struct _RESET_EVENT_INFO
{
   int   event;    // Reset Event that needs to be reported
   int   node;     // node that is resetting or disconnected

} RESET_EVENT_INFO, *PRESET_EVENT_INFO;

// ******************************************************
//  MEMORY ACCESSABLE BY ALL PROCESSES ACCESSING THE SP *
// ******************************************************
typedef struct _SP_SHARED_MEM
{
   int              InitialLoad;
   int              InitialLog;                 // True until logged message for successful load
   int              Resetting;                  // Reset in progress flag
   int              CallersHandle;              // Callers Handle
   int              CallersNode;                // Callers node
   int              SpCount;                    // # of Service Processors
   int              BuffLen;                    // Length of data to be copied
   int              CmdRespReceived;            // Flag indicating to uwiapi that there is a command response available
   pid_t            IbmasmsPid;                 // IBMASMs PID
   pid_t            ChildsPid;                  // Childs PID
   pid_t            CallersPid;                 // Callers PID
   pid_t            PadPid;                     // Callers PID
   RESET_EVENT_INFO ResetEventInfo;             // PHR_225158
   SP_HANDLE_LIST   HandleList[MAX_HANDLES];    // Queue of locally issued handles
   SP_CALLER_LIST   CallerList[MAX_HANDLES];    // Queue of Callers
   SP_DEV_NODE      SPDevNode[MAX_SP_NODES];    // SP Node info - H/W instance
   unsigned char    Buffer[MAX_BUFF_SIZE];      // Command/Receive buffer
   unsigned char    CmdBuff[16];                // Holds the callers cmd request

   // Start of PHR_LINK    DYNAMIC LINK TO LIBUSB
   void *pLib;

   SP_USB_INIT                sp_usb_init;
   SP_USB_CLOSE               sp_usb_close;
   SP_USB_FIND_BUSSES         sp_usb_find_busses;
   SP_USB_FIND_DEVICES        sp_usb_find_devices;
   SP_USB_BULK_WRITE          sp_usb_bulk_write;
   SP_USB_BULK_READ           sp_usb_bulk_read;
   SP_USB_CONTROL_MSG         sp_usb_control_msg;
   SP_USB_SET_CONFIGURATION   sp_usb_set_configuration;
   SP_USB_CLAIM_INTERFACE     sp_usb_claim_interface;
   SP_USB_RELEASE_INTERFACE   sp_usb_release_interface;
   SP_USB_SET_ALTINTERFACE    sp_usb_set_altinterface;
   SP_USB_RESETEP             sp_usb_resetep;
   SP_USB_CLEAR_HALT          sp_usb_clear_halt;
   SP_USB_RESET               sp_usb_reset;
   SP_USB_GET_BUSSES          sp_usb_get_busses;
   SP_USB_OPEN                sp_usb_open;
   SP_USB_DEVICE              sp_usb_device;
   SP_USB_DETACH              sp_usb_detach_kernel_driver_np;
   // End of PHR_LINK
   
   // End of PHR_LINK

} SP_SHARED_MEM, *PSP_SHARED_MEM;

